/*
 * BRLTTY - Access software for Unix for a blind person
 *          using a soft Braille terminal
 *
 * Copyright (C) 1995-1998 by The BRLTTY Team, All rights reserved.
 *
 *    Nicolas Pitre <nico@cam.org>
 *    Stphane Doyon <s.doyon@videotron.ca>
 *    Nikhil Nair <nn201@cus.cam.ac.uk>
 * Dynamic loading added by
 *    Roger Butenuth <butenuth@uni-paderborn.de>
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 *
 * This software is maintained by Nicolas Pitre <nico@cam.org>.
 */

/*
 * Alva_ABT3/brl.c - Braille display library for Alva ABT3xx series
 * Copyright (C) 1995-1996 by Nicolas Pitre <nico@cam.org>
 * See the GNU Public license for details in the ../COPYING file
 *
 * $Id: brl.c,v 1.10 1999/03/09 21:43:11 butenuth Exp $
 */

#define __EXTENSIONS__	/* for termios.h */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/termios.h>
#include <string.h>
#include <assert.h>

#include "../brl.h"
#include "../misc.h"
#include "../config.h"
#include "../config_file.h"


#define BAUDRATE B9600

/*
 * typematic settings
 */
#define TYPEMATIC_DELAY  10	/* number of cycles before a key is repeated */
#define TYPEMATIC_REPEAT  2	/* number of cycles between each key repeat */

/*
 * Delay in miliseconds between forced full refresh of the display.
 * This is to minimize garbage effects due to noise on the serial line.
 */
#define REFRESH_RATE 1000

/*
 * Device Identification Numbers (not to be changed).
 */
typedef enum {
    ABT320 = 0,
    ABT340 = 1,
    ABT34D = 2,
    ABT380 = 3,			/* ABT340 Desktop */
    ABT38D = 4,			/* ABT380 Twin Space */
    ABT_MODEL_COUNT = 5,
    ABT_AUTO = 6,		/* for new firmware only */
} abt_model_t;


#define MAX_STCELLS	5	/* highest number of status cells */

/*
 * Braille display parameters
 */
typedef struct {
    char *name;
    int  cols;
    int  status_cells;
} brl_params_t;

brl_params_t models[] = {
    { "ABT320",            20, 3 },
    { "ABT340",            40, 3 },
    { "ABT340 Desktop",    40, 5 },
    { "ABT380",            80, 5 },
    { "ABT380 Twin Space", 80, 5 },
    { NULL, 0, 0 },
    { "ABT_AUTO", 0, 0 }
};
/*
 * This is the brltty braille mapping standard to Alva's mapping table.
 */
static char TransTable[256] = {
    0x00, 0x01, 0x08, 0x09, 0x02, 0x03, 0x0A, 0x0B,
    0x10, 0x11, 0x18, 0x19, 0x12, 0x13, 0x1A, 0x1B,
    0x04, 0x05, 0x0C, 0x0D, 0x06, 0x07, 0x0E, 0x0F,
    0x14, 0x15, 0x1C, 0x1D, 0x16, 0x17, 0x1E, 0x1F,
    0x20, 0x21, 0x28, 0x29, 0x22, 0x23, 0x2A, 0x2B,
    0x30, 0x31, 0x38, 0x39, 0x32, 0x33, 0x3A, 0x3B,
    0x24, 0x25, 0x2C, 0x2D, 0x26, 0x27, 0x2E, 0x2F,
    0x34, 0x35, 0x3C, 0x3D, 0x36, 0x37, 0x3E, 0x3F,
    0x40, 0x41, 0x48, 0x49, 0x42, 0x43, 0x4A, 0x4B,
    0x50, 0x51, 0x58, 0x59, 0x52, 0x53, 0x5A, 0x5B,
    0x44, 0x45, 0x4C, 0x4D, 0x46, 0x47, 0x4E, 0x4F,
    0x54, 0x55, 0x5C, 0x5D, 0x56, 0x57, 0x5E, 0x5F,
    0x60, 0x61, 0x68, 0x69, 0x62, 0x63, 0x6A, 0x6B,
    0x70, 0x71, 0x78, 0x79, 0x72, 0x73, 0x7A, 0x7B,
    0x64, 0x65, 0x6C, 0x6D, 0x66, 0x67, 0x6E, 0x6F,
    0x74, 0x75, 0x7C, 0x7D, 0x76, 0x77, 0x7E, 0x7F,
    0x80, 0x81, 0x88, 0x89, 0x82, 0x83, 0x8A, 0x8B,
    0x90, 0x91, 0x98, 0x99, 0x92, 0x93, 0x9A, 0x9B,
    0x84, 0x85, 0x8C, 0x8D, 0x86, 0x87, 0x8E, 0x8F,
    0x94, 0x95, 0x9C, 0x9D, 0x96, 0x97, 0x9E, 0x9F,
    0xA0, 0xA1, 0xA8, 0xA9, 0xA2, 0xA3, 0xAA, 0xAB,
    0xB0, 0xB1, 0xB8, 0xB9, 0xB2, 0xB3, 0xBA, 0xBB,
    0xA4, 0xA5, 0xAC, 0xAD, 0xA6, 0xA7, 0xAE, 0xAF,
    0xB4, 0xB5, 0xBC, 0xBD, 0xB6, 0xB7, 0xBE, 0xBF,
    0xC0, 0xC1, 0xC8, 0xC9, 0xC2, 0xC3, 0xCA, 0xCB,
    0xD0, 0xD1, 0xD8, 0xD9, 0xD2, 0xD3, 0xDA, 0xDB,
    0xC4, 0xC5, 0xCC, 0xCD, 0xC6, 0xC7, 0xCE, 0xCF,
    0xD4, 0xD5, 0xDC, 0xDD, 0xD6, 0xD7, 0xDE, 0xDF,
    0xE0, 0xE1, 0xE8, 0xE9, 0xE2, 0xE3, 0xEA, 0xEB,
    0xF0, 0xF1, 0xF8, 0xF9, 0xF2, 0xF3, 0xFA, 0xFB,
    0xE4, 0xE5, 0xEC, 0xED, 0xE6, 0xE7, 0xEE, 0xEF,
    0xF4, 0xF5, 0xFC, 0xFD, 0xF6, 0xF7, 0xFE, 0xFF
};



/*
 * Global variables
 */
static int brl_fd;		/* file descriptor for Braille display */
static struct termios oldtio;	/* old terminal settings */
static unsigned char *rawdata;	/* translated data to send to Braille */
static unsigned char *prevdata;	/* previously sent raw data */
static unsigned char StatusCells[MAX_STCELLS];	/* to hold status info */
static unsigned char PrevStatus[MAX_STCELLS];	/* to hold previous status */
static brl_params_t *model_p;	/* points to terminal model config struct */
static int ReWrite = 0;		/* 1 if display need to be rewritten */


/*
 * NB: The first 7 key values are the same as those returned by the
 * old firmware, so they can be used directly from the input stream as
 * make and break sequence already combined... not to be changed.
 */
#define KEY_PROG 	0x008	/* the PROG key */
#define KEY_HOME 	0x004	/* the HOME key */
#define KEY_CURSOR 	0x002	/* the CURSOR key */
#define KEY_UP 		0x001	/* the UP key */
#define KEY_LEFT 	0x010	/* the LEFT key */
#define KEY_RIGHT 	0x020	/* the RIGHT key */
#define KEY_DOWN 	0x040	/* the DOWN key */
#define KEY_CURSOR2 	0x080	/* the CURSOR2 key */
#define KEY_HOME2 	0x100	/* the HOME2 key */
#define KEY_PROG2 	0x200	/* the PROG2 key */
#define KEY_ROUTING	0x400	/* cursor routing key set */

/*
 * those keys are not supposed to be combined, so their corresponding 
 * values are not bit exclusive between them.
 */
#define KEY_ROUTING_A	0x1000	/* first cur routing over status display */
#define KEY_ROUTING_B	0x2000	/* second cur routing over status disp. */
#define KEY_ROUTING_C	0x3000	/* ... */
#define KEY_ROUTING_D	0x4000
#define KEY_ROUTING_E	0x5000
#define KEY_ROUTING_F	0x6000

/* first cursor routing offset on main display (old firmware only) */
#define KEY_ROUTING_OFFSET 168

/* Index for new firmware protocol */
int OperatingKeys[10] = {
    KEY_PROG, KEY_HOME, KEY_CURSOR,
    KEY_UP, KEY_LEFT, KEY_RIGHT, KEY_DOWN,
    KEY_CURSOR2, KEY_HOME2, KEY_PROG2
};

int StatusKeys[6] = {
    KEY_ROUTING_A, KEY_ROUTING_B, KEY_ROUTING_C,
    KEY_ROUTING_D, KEY_ROUTING_E, KEY_ROUTING_F
};

char BRL_ID[] = "\033ID=";
#define DIM_BRL_ID 4



static int  get_config_vars(brl_drv_t *brl_drv);
static void closebrl(struct brl_disp *disp);
static void writebrl(struct brl_disp *disp);
static int  readbrl(struct brl_disp *disp, int flag);
static void setbrlstat(struct brl_disp *disp);

static int old_get_key(unsigned int *Keys, unsigned int *Pos);
static int new_get_key(unsigned int *Keys, unsigned int *Pos);



/*
 * Parameters that can be set from the configuration file.
 */
static abt_model_t model;
static char *device_name;
static int (*get_abt_key)(unsigned int *Keys, unsigned int *Pos) = new_get_key;

/*
 * ----------------------------------------------------------------------
 * Called when library is loaded.
 * ----------------------------------------------------------------------
 */
void _init(void)
{
    fprintf(stderr, "Alva serial driver loaded\n");
}


/*
 * ----------------------------------------------------------------------
 * Called before library is unloaded.
 * ----------------------------------------------------------------------
 */
void _fini(void)
{
    ;
}


/*
 * ----------------------------------------------------------------------
 * This function is called before `deamonize', so we can use printf
 * for diagnostic output.
 * Return 0 on success.
 * ----------------------------------------------------------------------
 */
int initbrl(brl_disp_t *disp, brl_drv_t *brl_drv)
{
    struct termios newtio;	/* new terminal settings */
    unsigned char  buffer[DIM_BRL_ID + 1];
    int            res;

    /*
     * Read the configuration variables we are interested in and store
     * results in global variables.
     */
    res = get_config_vars(brl_drv);
    if (res != 0)
	return res;

    disp->disp = NULL;
    rawdata = prevdata = NULL;		/* clear pointers */
    /*
     * Open the Braille display device for read/write.
     */
    brl_fd = open (device_name, O_RDWR | O_NOCTTY);
    if (brl_fd < 0)
	return 1;
    tcgetattr (brl_fd, &oldtio);	/* save current settings */
    /*
     * Set flow control and 8n1, enable reading
     */
    newtio.c_cflag =  CRTSCTS | CS8 | CLOCAL | CREAD;
    /*
     * Ignore bytes with parity errors and make terminal raw and dumb
     */
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;		/* raw output */
    newtio.c_lflag = 0;		/* don't echo or generate signals */
    newtio.c_cc[VMIN] = 0;	/* set nonblocking read */
    newtio.c_cc[VTIME] = 0;
    /*
     * autodetecting ABT model
     */
    do {
	/*
	 * to force DTR off
	 */
	cfsetispeed(&newtio, B0);
	cfsetospeed(&newtio, B0);
	tcsetattr(brl_fd, TCSANOW, &newtio);	/* activate new settings */
	tcflush(brl_fd, TCIOFLUSH);	/* clean line */
	delay(400);
	/*
	 * DTR back on
	 */
	cfsetispeed(&newtio, BAUDRATE);
	cfsetospeed(&newtio, BAUDRATE);
	tcsetattr(brl_fd, TCSANOW, &newtio);	/* activate new settings */
	delay(600);		/* give time to send ID string */
	/*
	 * The 2 next lines can be commented out
	 * to try autodetect once anyway
	 */
	if (model != ABT_AUTO)
	    break;
	if (read(brl_fd, &buffer, DIM_BRL_ID + 1) == DIM_BRL_ID + 1) {
	    if (!strncmp(buffer, BRL_ID, DIM_BRL_ID))
		model = buffer[DIM_BRL_ID];
	}
    } while (model == ABT_AUTO);
    if (model >= ABT_MODEL_COUNT || model < 0)
	return 1;
    /*
     * Set model params...
     */
    model_p = &models[model];
    disp->x = model_p->cols;		/* initialise size of display */
    disp->y = 1;
    disp->stat_mode = ST_AlvaStyle;
    disp->s = model_p->status_cells;
    /*
     * Set method pointers.
     */
    disp->closebrl   = closebrl;
    disp->writebrl   = writebrl;
    disp->readbrl    = readbrl;
    disp->setbrlstat = setbrlstat;
    /*
     * Allocate space for buffers.
     */
    disp->disp = malloc (disp->x * disp->y);
    disp->stat = malloc(disp->s);
    rawdata  = malloc (disp->x * disp->y);
    prevdata = malloc (disp->x * disp->y);
    if (!disp->disp || !disp->stat || !rawdata || !prevdata) {
	if (disp->disp)
	    free (disp->disp);
	if (disp->stat)
	    free (disp->stat);
	if (rawdata)
	    free (rawdata);
	if (prevdata)
	    free (prevdata);
	return 1;
    } else {
	ReWrite = 1;		/* To write whole display at first time */
	return 0;
    }
}


/*
 * ----------------------------------------------------------------------
 * model, device_name, get_abt_key
 * ----------------------------------------------------------------------
 */
static int  get_config_vars(brl_drv_t *brl_drv)
{
    int      res = 0;
    symbol_t *sym;
    char     *str;
    int      i;

    device_name = NULL;
    sym = lookup_symbol(brl_drv->scope, "device");
    if (sym != NULL) {
	if (sym->type != SY_STRINGVAR)
	    fprintf(stderr, "\"device\" must be string\n");
	else
	    device_name = sym->u.stringval;
    }
    if (device_name == NULL)
	device_name = "/dev/ttyS0";

    model = ABT_AUTO;
    sym = lookup_symbol(brl_drv->scope, "model");
    if (sym != NULL) {
	if (sym->type != SY_STRINGVAR)
	    fprintf(stderr, "\"model\" must be string\n");
	else {
	    str = sym->u.stringval;
	    for (i = 0; models[i].name != NULL; i++) {
		if (strcasecmp(str, models[i].name) == 0) {
		    model = i;
		    break;
		}
	    }
	}
    }
    
    get_abt_key = new_get_key;
    sym = lookup_symbol(brl_drv->scope, "firmware");
    if (sym != NULL) {
	if (sym->type != SY_STRINGVAR)
	    fprintf(stderr, "\"firmware\" must be string\n");
	else {
	    if (strcasecmp("old", sym->u.stringval) == 0)
		get_abt_key = old_get_key;
	    else if (strcasecmp("new", sym->u.stringval) == 0)
		get_abt_key = new_get_key;
	    else
		fprintf(stderr, "\"firmware\" must be \"old\" or \"new\","
			" not \"%s\"\n", sym->u.stringval);
	}
    }
    
    if (get_abt_key == old_get_key && model == ABT_AUTO) {
	fprintf(stderr,
		"Warning: Old firmware does not support auto detection of"
		"hardware, assuming to new firmware\n");
	get_abt_key = new_get_key;
    }
    
    printf("device   = \"%s\"\n", device_name);
    if (get_abt_key == old_get_key)
	printf("firmware = \"old\" (pre 95/0401)\n");
    else
	printf("firmware = \"new\" (post 95/0401)\n");
    printf("model    = \"%s\"\n", models[model].name);

    return res;
}


/*
 * ----------------------------------------------------------------------
 * Reset terminal parameters, free buffers, close serial port.
 * ----------------------------------------------------------------------
 */
static void closebrl(struct brl_disp *disp)
{
    free(disp->disp);
    free(disp->stat);
    free(rawdata);
    free(prevdata);
    tcsetattr(brl_fd, TCSANOW, &oldtio); /* restore terminal settings */
    close(brl_fd);
}


/*
 * ----------------------------------------------------------------------
 * Write a string to the braille display.
 * Coordinates start at zero with the status cells.
 * Return 0 on success, 1 on error.
 * ----------------------------------------------------------------------
 */
static int WriteToBrlDisplay(int start, int len, unsigned char *data)
{
    static char prefix[5] = "\r\033B  ";

    prefix[3] = start;
    prefix[4] = len;

    if (write(brl_fd, prefix, 5) == 5)
	if (write(brl_fd, data, len) == len)
	    if (write(brl_fd, "\r", 1) == 1) {
		return 0;
	    }
    
    return 1;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static void writebrl(struct brl_disp *disp)
{
    int i, j, k;
    static int Timeout = 0;

    if (ReWrite || ++Timeout > (REFRESH_RATE/DELAY_TIME)) {
	ReWrite = Timeout = 0;
	/*
	 * We rewrite the whole display
	 */
	i = 0;
	j = model_p->cols;
    } else {
	/*
	 * We update only the display part that has been changed
	 */
	i = 0;
	while ((disp->disp[i] == prevdata[i]) && (i < model_p->cols))
	    i++;
	j = model_p->cols - 1;
	while ((disp->disp[j] == prevdata[j]) && (j >= i))
	    j--;
	j++;
    }
    if (i < j) {
	/*
	 * there is something different
	 */
	for (k = 0; k < j - i; k++) {
	    prevdata[i + k] = disp->disp[i + k];
	    rawdata[k] = TransTable[disp->disp[i + k]];
	}
	WriteToBrlDisplay(model_p->status_cells + i, j - i, rawdata);
    }
}


/*
 * ----------------------------------------------------------------------
 * Update status cells on braille display
 * ----------------------------------------------------------------------
 */
static void setbrlstat(struct brl_disp *disp)
{
    int i;

    if (memcmp(disp->stat, PrevStatus, disp->s)) {
	/*
	 * only if it changed
	 */
	for (i = 0; i < disp->s; i++) {
	    PrevStatus[i] = disp->stat[i];
	    StatusCells[i] = TransTable[disp->stat[i]];
	}
	WriteToBrlDisplay(0, disp->s, StatusCells);
    }
}


/*
 * ----------------------------------------------------------------------
 * Version for old firmware
 * ----------------------------------------------------------------------
 */
static int old_get_key(unsigned int *Keys, unsigned int *Pos)
{
    unsigned char c;
    
    while (read(brl_fd, &c, 1)) {
	if ((c >= (KEY_ROUTING_OFFSET + model_p->cols)) &&
	    (c < (KEY_ROUTING_OFFSET + model_p->cols + 6))) {
	    /*
	     * make for Status keys of Touch Cursor
	     */
	    *Keys |= StatusKeys[c - (KEY_ROUTING_OFFSET + model_p->cols)];
	} else if (c >= KEY_ROUTING_OFFSET) {
	    /*
	     * make for display keys of Touch cursor
	     */
	    *Pos = c - KEY_ROUTING_OFFSET;
	    *Keys |= KEY_ROUTING;
	} else {
	    /*
	     * check comments where KEY_xxxx are defined
	     */
	    *Keys = c;
	}
	return 1;
    }

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * Version for new firmware
 * ----------------------------------------------------------------------
 */
static int new_get_key(unsigned int *Keys, unsigned int *Pos)
{
    unsigned char c;
    static int KeyGroup = 0;
    
    while (read(brl_fd, &c, 1)) {
	switch (KeyGroup) {
	case 0x71:	/* Operating keys and Status keys of Touch Cursor */
	    /* make keys */
	    if (c <= 0x09)
		*Keys |= OperatingKeys[c];
	    else if ((c >= 0x20) && (c <= 0x25))
		*Keys |= StatusKeys[c - 0x20];
	    /* break keys */
	    else if ((c >= 0x80) && (c <= 0x89))
		*Keys &= ~OperatingKeys[c - 0x80];
	    else if ((c >= 0xA0) && (c <= 0xA5))
		*Keys &= ~StatusKeys[c - 0xA0];
	    else
		*Keys = 0;
	    KeyGroup = 0;
	    return 1;

	case 0x72:		/* Display keys of Touch Cursor */
	    /* make keys */
	    if (c <= 0x5F) {
		/* make */
		*Pos = c;
		*Keys |= KEY_ROUTING;
	    }
	    /* break keys */
	    else
		*Keys &= ~KEY_ROUTING;
	    KeyGroup = 0;
	    return 1;
	default:
	    /*
	     * Key group selection
	     */
	    if ((c == 0x71) || (c == 0x72))
		KeyGroup = c;
	    else {
		/*
		 * Let's reset the keys and rewrite the display
		 */
		*Keys = 0;
		ReWrite = 1;
	    }
	}
	
    }

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int readbrl(struct brl_disp *disp, int type)
{
    int ProcessKey, res = EOF;
    static unsigned int RoutingPos = 0;
    static unsigned int ActualKeys = 0, LastKeys = 0, ReleasedKeys = 0;
    static int Typematic = 0, KeyDelay = 0, KeyRepeat = 0;
    
    if (!(ProcessKey = (*get_abt_key)(&ActualKeys, &RoutingPos))) {
	if (Typematic) {
	    /*
	     * a key is being hold
	     */
	    if (KeyDelay < TYPEMATIC_DELAY)
		KeyDelay++;
	    else if (KeyRepeat < TYPEMATIC_REPEAT)
		KeyRepeat++;
	    else {
		/*
		 * It's time to issue the command again
		 */
		ActualKeys = LastKeys;
		LastKeys = 0;
		KeyRepeat = 0;
		ProcessKey = 1;
	    }
	}
    } else {
	/*
	 * A new key is being pressed/released, we clear the typematic dounters
	 */
	Typematic = KeyDelay = KeyRepeat = 0;
    }
    
    if (ProcessKey) {
	if (ActualKeys > LastKeys) {
	    /*
	     * These are the keys that should be processed when pressed
	     */
	    LastKeys = ActualKeys;	/* we keep it until it is released */
	    ReleasedKeys = 0;
	    switch (ActualKeys) {
	    case KEY_HOME | KEY_UP:
		res = CMD_TOP;
		break;
	    case KEY_HOME | KEY_DOWN:
		res = CMD_BOT;
		break;
	    case KEY_UP:
		res = CMD_LNUP;
		Typematic = 1;
		break;
	    case KEY_CURSOR | KEY_UP:
		res = CMD_ATTRUP;
		break;
	    case KEY_DOWN:
		res = CMD_LNDN;
		Typematic = 1;
		break;
	    case KEY_CURSOR | KEY_DOWN:
		res = CMD_ATTRDN;
		break;
	    case KEY_LEFT:
		res = CMD_FWINLT;
		break;
	    case KEY_HOME | KEY_LEFT:
		res = CMD_LNBEG;
		break;
	    case KEY_CURSOR | KEY_LEFT:
		res = CMD_HWINLT;
		break;
	    case KEY_PROG | KEY_LEFT:
		res = CMD_CHRLT;
		Typematic = 1;
		break;
	    case KEY_RIGHT:
		res = CMD_FWINRT;
		break;
	    case KEY_HOME | KEY_RIGHT:
		res = CMD_LNEND;
		break;
	    case KEY_PROG | KEY_RIGHT:
		res = CMD_CHRRT;
		Typematic = 1;
		break;
	    case KEY_CURSOR | KEY_RIGHT:
		res = CMD_HWINRT;
		break;
	    case KEY_HOME | KEY_CURSOR | KEY_UP:
		res = CMD_PRDIFLN;
		break;
	    case KEY_HOME | KEY_CURSOR | KEY_DOWN:
		res = CMD_NXDIFLN;
		break;
	    case KEY_HOME | KEY_CURSOR | KEY_LEFT:
		res = CMD_MUTE;
		break;
	    case KEY_HOME | KEY_CURSOR | KEY_RIGHT:
		res = CMD_SAY;
		break;
	    case KEY_PROG | KEY_CURSOR:
		res = CMD_CONFMENU;
		break;
	    case KEY_PROG | KEY_DOWN:
		res = CMD_FREEZE;
		break;
	    case KEY_PROG | KEY_UP:
		res = CMD_INFO;
		break;
	    case KEY_ROUTING_A:
		res = CMD_CAPBLINK;
		break;
	    case KEY_ROUTING_B:
		res = CMD_CSRVIS;
		break;
	    case KEY_ROUTING_C:
		res = CMD_CSRBLINK;
		break;
	    case KEY_CURSOR | KEY_ROUTING_A:
		res = CMD_SIXDOTS;
		break;
	    case KEY_CURSOR | KEY_ROUTING_B:
		res = CMD_CSRSIZE;
		break;
	    case KEY_CURSOR | KEY_ROUTING_C:
		res = CMD_SLIDEWIN;
		break;
	    case KEY_PROG | KEY_HOME | KEY_LEFT:
		res = CMD_CUT_BEG;
		break;
	    case KEY_PROG | KEY_HOME | KEY_RIGHT:
		res = CMD_CUT_END;
		break;
	    case KEY_ROUTING:
		/* normal Cursor routing keys */
		res = CR_ROUTEOFFSET + RoutingPos;
		break;
	    case KEY_PROG | KEY_ROUTING:
		/* marking beginning of block */
		res = CR_BEGBLKOFFSET + RoutingPos;
		break;
	    case KEY_HOME | KEY_ROUTING:
		/* marking end of block */
		res = CR_ENDBLKOFFSET + RoutingPos;
		break;
	    case KEY_PROG | KEY_HOME | KEY_DOWN:
		res = CMD_PASTE;
		break;
	    }
	} else {
	    /*
	     * These are the keys that should be processed when released
	     */
	    if (!ReleasedKeys) {
		ReleasedKeys = LastKeys;
		switch (ReleasedKeys) {
		case KEY_HOME:
		    res = CMD_TOP_LEFT;
		    break;
		case KEY_CURSOR:
		    res = CMD_HOME;
		    ReWrite = 1;	/* force rewrite of whole display */
		    break;
		case KEY_PROG:
		    res = CMD_HELP;
		    /*res = CMD_SAY;*/
		    break;
		case KEY_PROG | KEY_HOME:
		    res = CMD_DISPMD;
		    break;
		case KEY_HOME | KEY_CURSOR:
		    res = CMD_CSRTRK;
		    break;
		}
	    }
	    LastKeys = ActualKeys;
	    if (!ActualKeys)
		ReleasedKeys = 0;
	}
    }

    return res;
}
