/*
 * BRLTTY - Access software for Unix for a blind person
 *          using a soft Braille terminal
 *
 * Copyright (C) 1995-1998 by The BRLTTY Team, All rights reserved.
 *
 *    Nicolas Pitre <nico@cam.org>
 *    Stphane Doyon <s.doyon@videotron.ca>
 *    Nikhil Nair <nn201@cus.cam.ac.uk>
 * Dynamic loading added by
 *    Roger Butenuth <butenuth@uni-paderborn.de>
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 */

/* CombiBraille/brl.c - Braille display library
 * For Tieman B.V.'s CombiBraille (serial interface only)
 * Maintained by Nikhil Nair <nn201@cus.cam.ac.uk>
 * $Id: brl.c,v 1.2 1999/03/09 17:56:05 butenuth Exp $
 */

#define __EXTENSIONS__  /* for termios.h */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/termios.h>
#include <string.h>

#include "../brl.h"
#include "../misc.h"
#include "../config_file.h"
#include "brlconf.h"
#include "tables.h"		/* for keybindings */

static unsigned char combitrans[256];	/* dot mapping table (output) */
static int brl_fd;		/* file descriptor for Braille display */
static char *device_name = NULL;
static unsigned char *prevdata;	/* previously received data */
static unsigned char status[5], oldstatus[5];  /* status cells - always five */
static unsigned char *rawdata;	/* writebrl() buffer for raw Braille data */
static short rawlen;		/* length of rawdata buffer */
struct termios oldtio;		/* old terminal settings */

/* Function prototypes: */
static int getbrlkey(void);	/* get a keystroke from the CombiBraille */

static int  get_config_vars(brl_drv_t *brl_drv);
static int  timeout_yet(int msec);
static void closebrl(struct brl_disp *disp);
static void writebrl(struct brl_disp *disp);
static int  readbrl(struct brl_disp *disp, int flag);
static void setbrlstat(struct brl_disp *disp);


/*
 * ----------------------------------------------------------------------
 * Called when library is loaded.
 * ----------------------------------------------------------------------
 */
void _init(void)
{
    fprintf(stderr, "CombiBraille serial driver loaded\n");
}


/*
 * ----------------------------------------------------------------------
 * Called before library is unloaded.
 * ----------------------------------------------------------------------
 */
void _fini(void)
{
    ;
}


/*
 * ----------------------------------------------------------------------
 * Main entry point, try to start everything.
 * ----------------------------------------------------------------------
 */
int initbrl(brl_disp_t *disp, brl_drv_t *brl_drv)
{
    struct termios newtio;	/* new terminal settings */
    short i, n, success;		/* loop counters, flags, etc. */
    unsigned char *init_seq = INIT_SEQ;	/* bytewise accessible copies */
    unsigned char *init_ack = INIT_ACK;
    unsigned char c;
    char id = -1;
    unsigned char standard[8] = {0, 1, 2, 3, 4, 5, 6, 7}; /* brltty standard */
    unsigned char Tieman[8]   = {0, 7, 1, 6, 2, 5, 3, 4}; /* Tieman standard */

    disp->disp = disp->stat = NULL; /* clear pointers */
    prevdata = rawdata = NULL;
    /*
     * Read the configuration variables we are interested in and store
     * results in global variables.
     */
    if (get_config_vars(brl_drv) != 0)
	goto failure;
    /*
     * Now open the Braille display device for random access
     */
    brl_fd = open (device_name, O_RDWR | O_NOCTTY);
    if (brl_fd < 0)
	goto failure;
    tcgetattr(brl_fd, &oldtio);	/* save current settings */
    
    /* Set bps, flow control and 8n1, enable reading */
    newtio.c_cflag = BAUDRATE | CRTSCTS | CS8 | CLOCAL | CREAD;

    /* Ignore bytes with parity errors and make terminal raw and dumb */
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;		/* raw output */
    newtio.c_lflag = 0;		/* don't echo or generate signals */
    newtio.c_cc[VMIN] = 0;	/* set nonblocking read */
    newtio.c_cc[VTIME] = 0;
    tcflush (brl_fd, TCIFLUSH);	/* clean line */
    tcsetattr (brl_fd, TCSANOW, &newtio); /* activate new settings */

    /* CombiBraille initialisation procedure: */
    success = 0;
    /* Try MAX_ATTEMPTS times, or forever if MAX_ATTEMPTS is 0: */
#if MAX_ATTEMPTS == 0
    while (!success)
#else
	for (i = 0; i < MAX_ATTEMPTS && !success; i++)
#endif
	{
	    if (init_seq[0])
		if (write (brl_fd, init_seq + 1, init_seq[0]) != init_seq[0])
		    continue;
	    timeout_yet(0);		/* initialise timeout testing */
	    n = 0;
	    do {
		delay (20);
		if (read (brl_fd, &c, 1) == 0)
		    continue;
		if (n < init_ack[0] && c != init_ack[1 + n])
		    continue;
		if (n == init_ack[0])
		    id = c, success = 1;
		n++;
	    } while (!timeout_yet(ACK_TIMEOUT) && n <= init_ack[0]);
	}
    if (!success) {
	tcsetattr (brl_fd, TCSANOW, &oldtio);
	goto failure;
    }

    disp->x = BRLCOLS(id);		/* initialise size of display */
    disp->y = BRLROWS;
    disp->s = 5;
    if (disp->x == -1)
	goto failure;
    /*
     * Set method pointers.
     */
    disp->closebrl   = closebrl;
    disp->writebrl   = writebrl;
    disp->readbrl    = readbrl;
    disp->setbrlstat = setbrlstat;
    /*
     * Allocate space for buffers
     */
    disp->disp = malloc(disp->x * disp->y);
    disp->stat = malloc(disp->s);
    prevdata = malloc(disp->x * disp->y);
    /*
     * rawdata has to have room for the pre- and post-data sequences,
     * the status cells, and escaped 0x1b's:
     */
    rawdata = malloc(20 + disp->x * disp->y * 2);
    if (!disp->disp || !disp->stat || !prevdata || !rawdata)
	goto failure;
    /*
     * Generate dot mapping table:
     */
    memset (combitrans, 0, 256);
    for (n = 0; n < 256; n++)
	for (i = 0; i < 8; i++)
	    if (n & 1 << standard[i])
		combitrans[n] |= 1 << Tieman[i];
    
    return 0;

failure:;
  if (disp->disp)
    free(disp->disp);
  if (prevdata)
    free(prevdata);
  if (rawdata)
    free(rawdata);
  if (brl_fd >= 0)
    close (brl_fd);
  disp->x = -1;

  return 1;
}


/*
 * ----------------------------------------------------------------------
 * Get device_name from config file.
 * ----------------------------------------------------------------------
 */
static int get_config_vars(brl_drv_t *brl_drv)
{
    int      res = 0;
    symbol_t *sym;

    device_name = NULL;
    sym = lookup_symbol(brl_drv->scope, "device");
    if (sym != NULL) {
	if (sym->type != SY_STRINGVAR)
	    fprintf(stderr, "\"device\" must be string\n");
	else
	    device_name = sym->u.stringval;
    }
    if (device_name == NULL)
	device_name = "/dev/ttyS0";

    return res;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int timeout_yet(int msec)
{
  static struct timeval tstart = { 0, 0 };
  struct timeval tnow;

  if (msec == 0) {	/* initialiseation */
      gettimeofday (&tstart, NULL);
      return 0;
  }
  gettimeofday (&tnow, NULL);
  return ((tnow.tv_sec * 1e6 + tnow.tv_usec) - \
	  (tstart.tv_sec * 1e6 + tstart.tv_usec) >= msec * 1e3);
}


/*
 * ----------------------------------------------------------------------
 * Reset terminal parameters, free buffers, close serial port.
 * ----------------------------------------------------------------------
 */
static void closebrl(struct brl_disp *disp)
{
    unsigned char *pre_data = PRE_DATA;	/* bytewise accessible copies */
    unsigned char *post_data = POST_DATA;
    unsigned char *close_seq = CLOSE_SEQ;
    
    rawlen = 0;
    if (pre_data[0]) {
	memcpy (rawdata + rawlen, pre_data + 1, pre_data[0]);
	rawlen += pre_data[0];
    }
    /* Clear the five status cells and the main display: */
    memset (rawdata + rawlen, 0, 5 + brl.x * brl.y);
    rawlen += 5 + brl.x * brl.y;
    if (post_data[0]) {
	memcpy (rawdata + rawlen, post_data + 1, post_data[0]);
	rawlen += post_data[0];
    }
    
    /* Send closing sequence: */
    if (close_seq[0]) {
	memcpy (rawdata + rawlen, close_seq + 1, close_seq[0]);
	rawlen += close_seq[0];
    }
    write (brl_fd, rawdata, rawlen);

    free (brl.disp);
    free (prevdata);
    free (rawdata);
    
#if 0
    tcsetattr (brl_fd, TCSANOW, &oldtio); /* restore terminal settings */
#endif
    close (brl_fd);
}


/*
 * ----------------------------------------------------------------------
 * Update status cells on braille display
 * ----------------------------------------------------------------------
 */
static void setbrlstat(struct brl_disp *disp)
{
    short i;
    
    /* Dot mapping: */
    for (i = 0; i < 5; status[i] = combitrans[disp->stat[i]], i++)
	;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static void writebrl(struct brl_disp *disp)
{
    short i;			/* loop counter */
    unsigned char *pre_data = PRE_DATA;	/* bytewise accessible copies */
    unsigned char *post_data = POST_DATA;

    /* Only refresh display if the data has changed: */
    if (memcmp(disp->disp, prevdata, disp->x * disp->y) ||
	memcmp (status, oldstatus, 5)) {
	/* Save new Braille data: */
	memcpy(prevdata, disp->disp, disp->x * disp->y);

	/* Dot mapping from standard to CombiBraille: */
	for (i = 0;
	     i < disp->x * disp->y;
	     disp->disp[i] = combitrans[disp->disp[i]], i++)
	    ;

	rawlen = 0;
	if (pre_data[0]) {
	    memcpy (rawdata + rawlen, pre_data + 1, pre_data[0]);
	    rawlen += pre_data[0];
	}
	for (i = 0; i < 5; i++) {
	    rawdata[rawlen++] = status[i];
	    if (status[i] == 0x1b)	/* CombiBraille hack */
		rawdata[rawlen++] = 0x1b;
	}
	for (i = 0; i < disp->x * disp->y; i++) {
	    rawdata[rawlen++] = disp->disp[i];
	    if (disp->disp[i] == 0x1b)	/* CombiBraille hack */
		rawdata[rawlen++] = disp->disp[i];
	}
	if (post_data[0]) {
	    memcpy (rawdata + rawlen, post_data + 1, post_data[0]);
	    rawlen += post_data[0];
	}
	write (brl_fd, rawdata, rawlen);
    }
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int readbrl(struct brl_disp *disp, int type)
{
    int c;
    static short status = 0;	/* cursor routing keys mode */

    c = getbrlkey();
    if (c == EOF)
	return EOF;
    c = type == TBL_CMD ? cmdtrans[c] : argtrans[c];
    if (c == 1 || c == 2) {
	status = c;
	return EOF;
    }
    if (c & 0x80)		/* cursor routing keys */
	switch (status) {
	case 0:			/* ordinary cursor routing */
	    return (c ^ 0x80) + CR_ROUTEOFFSET;
	case 1:			/* begin block */
	    status = 0;
	    return (c ^ 0x80) + CR_BEGBLKOFFSET;
	case 2:			/* end block */
	    status = 0;
	    return (c ^ 0x80) + CR_ENDBLKOFFSET;
	}
    status = 0;
    return c;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int getbrlkey(void)
{
    static short ptr = 0;		/* input queue pointer */
    static unsigned char q[4];	/* input queue */
    unsigned char c;		/* character buffer */
    
    while (read (brl_fd, &c, 1)) {
	if (ptr == 0 && c != 27)
	    continue;
	if (ptr == 1 && c != 'K' && c != 'C') {
	    ptr = 0;
	    continue;
	}
	q[ptr++] = c;
	if (ptr < 3 || (ptr == 3 && q[1] == 'K' && !q[2]))
	    continue;
	ptr = 0;
	if (q[1] == 'K')
	    return (q[2] ? q[2] : q[3] | 0x0060);
	return ((int) q[2] | 0x80);
    }
    return EOF;
}
