/*
 * ======================================================================
 * Synthesizer module for "rsynth".
 * Copyright (C) 1998-1999 Roger Butenuth <butenuth@uni-paderborn.de>
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 * For the latest version see:
 * http://www.uni-paderborn.de/cs/heiss/blinux/index.html
 * $Id: rsynth.c,v 1.1 1999/03/11 21:08:58 butenuth Exp $
 * ======================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include "synthesizer.h"

static int  s_close(synth_t *s);
static int  s_synth(synth_t *s, unsigned char *buffer);
static int  s_flush(synth_t *s);
static int  s_clear(synth_t *s);
static int  s_sync(synth_t *s, sync_par_t par, int timeout);
static int  s_get_param(struct synth_struct *s, synth_par_t par, int *value);
static int  s_set_param(struct synth_struct *s, synth_par_t par, int value);

typedef struct synth_state {
    int        param[S_MAX];
    int        initialized;
    const char *argv[2];
} synth_state_t;

static int start_program(const char *argv[], int *to, int *from, int *pid);

static synth_state_t english_state;

static FILE *to_fp = NULL;
static int  to_fd, from_fd, speech_pid;


static synth_t english_rsynth = {
    &english_state,
    &languages[LANG_BRITISH_ENGLISH],
    "Rsynth/British English",
    NULL,
    s_close,
    s_synth,
    s_flush,
    s_clear,
    s_sync,
    s_get_param,
    s_set_param,
};


/*
 * ----------------------------------------------------------------------
 * Called before library is loaded.
 * ----------------------------------------------------------------------
 */
void _init(void)
{
    fprintf(stderr, "rsynth loaded\n");
}


/*
 * ----------------------------------------------------------------------
 * Called before library is unloaded.
 * ----------------------------------------------------------------------
 */
void _fini(void)
{
}


/*
 * ----------------------------------------------------------------------
 * General open function.
 * Return handle on succes, NULL on error.
 * ----------------------------------------------------------------------
 */
synth_t *synth_open(const char *arg)
{
    int         r = 0;

    english_state.argv[0] = strdup(arg);
    assert(english_state.argv[0] != NULL);
    english_state.argv[1] = NULL;
    r = start_program(english_state.argv, &to_fd, &from_fd, &speech_pid);
    if (r != 0)
	return NULL;
    to_fp = fdopen(to_fd, "w");
    if (to_fp == NULL)
	return NULL;

    english_state.param[S_SPEED]  = 100;
    english_state.param[S_PITCH]  = 100;
    english_state.param[S_VOLUME] =  75;
    english_state.initialized = 1;
    
    return &english_rsynth;
}


/*
 * ----------------------------------------------------------------------
 * General close.
 * ----------------------------------------------------------------------
 */
static int s_close(synth_t *s)
{
    int   result = 0;
    int   status;
    pid_t pid;

    assert(s->state->initialized);
    s->state->initialized = 0;
    free((void*)english_state.argv[0]);
    fclose(to_fp);
    to_fp   = NULL;
    pid = waitpid(speech_pid, &status, 0);
    if (WIFEXITED(status))
	result = WEXITSTATUS(status);
    else if (WIFSIGNALED(status))
	result = WTERMSIG(status);
    else
	result = 1;

    return result;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int s_synth(struct synth_struct *s, unsigned char *buffer)
{
    assert(s->state->initialized);

    fwrite(buffer, 1, strlen(buffer), to_fp);
    fflush(to_fp);

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int s_flush(synth_t *s)
{
    assert(s->state->initialized);

    fprintf(to_fp, ".\n");
    fflush(to_fp);

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * The rsynth program does not support clearing the speech pipeline,
 * just kill and restart.
 * ----------------------------------------------------------------------
 */
static int s_clear(synth_t *s)
{
#if 0
    int r;
    int pid;
    int status;
    int fd;

    assert(s->state->initialized);

    fclose(to_fp);
    to_fp   = NULL;
    kill(speech_pid, SIGKILL);

    pid = waitpid(speech_pid, &status, 0);
    if (WIFEXITED(status))
	r = WEXITSTATUS(status);
    else if (WIFSIGNALED(status))
	r = WTERMSIG(status);
    else
	r = 1;

    fd = open("/dev/dsp", O_WRONLY);
    if (fd >= 0) {
	ioctl(fd, SNDCTL_DSP_RESET);
	close(fd);
    }

    r = start_program(english_state.argv, &to_fd, &from_fd, &speech_pid);
    if (r != 0)
	return r;
    to_fp = fdopen(to_fd, "w");
    if (to_fp == NULL)
	return 1;
#endif    
    return 0;
}


/*
 * ----------------------------------------------------------------------
 * A sync mark is set by sending a string like "@I1\n" (other numbers
 * possible). When it is reached, a line "@I1\n" is sent back.
 * Return 1 when we have reached the mark, 0 when not, 2 on timeout,
 * and -1 on error.
 * ----------------------------------------------------------------------
 */
static int s_sync(synth_t *s, sync_par_t par, int timeout)
{
    int res;
    assert(s->state->initialized);

    res = -1;			/* not implemented */

    return res;
}


/*
 * ----------------------------------------------------------------------
 * Get a synthesizer parameter.
 * ----------------------------------------------------------------------
 */
static int s_get_param(struct synth_struct *s, synth_par_t par, int *value)
{
    if (par >= 0 && par < S_MAX) {
	*value = s->state->param[par];
	return 0;
    } else
	return 1;
}


/*
 * ----------------------------------------------------------------------
 * Set a parameter of the synthesizer. There is one problem: When we
 * clear immediately after issuing a command, the command may be lost.
 * For this reason, record the time of the command and sleep in the
 * clear function when there was a command in the last second.
 * Setting is only forwarded to the synthesizer if the selected 
 * synthesizer is the currently active one.
 * ----------------------------------------------------------------------
 */
static int s_set_param(struct synth_struct *s, synth_par_t par, int value)
{
    if (par >= 0 && par < S_MAX) {
	s->state->param[par] = value;
	return 0;
    } else
	return 1;
}


/*
 * ----------------------------------------------------------------------
 * Start a program (int the background) and connect it with two pipes
 * to the caller:
 *  *to from caller to program
 *  *from from programm to caller
 * Process id of the program is returned in *pid.
 * ----------------------------------------------------------------------
 */
static int start_program(const char *argv[], int *to, int *from, int *pid)
{
    int to_pipe[2], from_pipe[2];
    int i;

    if (pipe(to_pipe) < 0 || pipe(from_pipe) < 0) {
	perror("create pipes");
	return -1;
    }
    *pid = fork();
    assert(*pid >= 0);
    if (*pid == 0) {		/* child */
	close(from_pipe[0]);
	close(to_pipe[1]);
	if (dup2(to_pipe[0], 0) < 0 || dup2(from_pipe[1], 1) < 0) {
	    perror("dup2 on pipes failed");
	    kill(getppid(), SIGKILL);
	    exit(1);
	}
	close(from_pipe[1]);
	close(to_pipe[0]);
	execvp(argv[0], (char *const*)argv);	/* should never return */
	fprintf(stderr, "argv[0] = \"%s\"\n", argv[0]);
	perror("exec failed");
	fprintf(stderr, "command and args:");
	for (i = 0; argv[i] != NULL; i++)
	    fprintf(stderr, " %s ", argv[i]);
	fprintf(stderr, "\n");
	kill(getppid(), SIGKILL);
	exit(1);
    } else {			/* parent */
	*to   = to_pipe[1];
	*from = from_pipe[0];
	close(to_pipe[0]);
	close(from_pipe[1]);
    }

    return 0;
}
