/*
 * ======================================================================
 * Synthesizer module for "Softspeech", a wrapper around MBROLA,
 * HADIFIX (german), and FREEPHONE (english).
 * Copyright (C) 1998-1999 Roger Butenuth <butenuth@uni-paderborn.de>
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 * For the latest version see:
 * http://www.uni-paderborn.de/cs/heiss/blinux/index.html
 * $Id: softspeech.c,v 1.12 1999/03/11 21:08:58 butenuth Exp $
 * ======================================================================
 */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "synthesizer.h"

static int  s_close(synth_t *s);
static int  s_synth(synth_t *s, unsigned char *buffer);
static int  s_flush(synth_t *s);
static int  s_clear(synth_t *s);
static int  s_sync(synth_t *s, sync_par_t par, int timeout);
static int  s_get_param(struct synth_struct *s, synth_par_t par, int *value);
static int  s_set_param(struct synth_struct *s, synth_par_t par, int value);

#define SPEECH_DIR "/opt/speech"
static char softspeech_path[] = "/opt/speech/speech";

typedef struct synth_state {
    int  param[S_MAX];
    int  initialized;
} synth_state_t;

static int start_program(char *argv[], int *to, int *from, int *pid);

static synth_state_t german_state;
static synth_state_t english_state;

static FILE *to_fp = NULL;
static int  to_fd, from_fd, speech_pid;
static int  fp_ref = 0;
static int  current_language = -1;

static int sync_mark_no = 0;	/* currently used number */
static struct timeval mark;	/* time the mark has been set */
static struct timeval last_command_time; /* set when a command is issued */

static synth_t german_softspeech = {
    &german_state,
    &languages[LANG_GERMAN],
    "Softspeech/German",
    NULL,
    s_close,
    s_synth,
    s_flush,
    s_clear,
    s_sync,
    s_get_param,
    s_set_param,
};

static synth_t english_softspeech = {
    &english_state,
    &languages[LANG_BRITISH_ENGLISH],
    "Softspeech/British English",
    NULL,
    s_close,
    s_synth,
    s_flush,
    s_clear,
    s_sync,
    s_get_param,
    s_set_param,
};


/*
 * ----------------------------------------------------------------------
 * Called before library is loaded.
 * ----------------------------------------------------------------------
 */
void _init(void)
{
    fprintf(stderr, "Softspeech loaded\n");
}


/*
 * ----------------------------------------------------------------------
 * Called before library is unloaded.
 * ----------------------------------------------------------------------
 */
void _fini(void)
{
}


/*
 * ----------------------------------------------------------------------
 * General open function for german and english synthesizer.
 * Second open increments refcount.
 * Return 0 on succes, 1 on error.
 * ----------------------------------------------------------------------
 */
synth_t *synth_open(const char *arg)
{
    char    *argv[2];
    synth_t *s;
    int     r = 0;

    gettimeofday(&last_command_time, NULL);
    if (fp_ref == 0) {
	argv[0] = softspeech_path;
	argv[1] = NULL;
	r = start_program(argv, &to_fd, &from_fd, &speech_pid);
	if (r != 0)
	    return NULL;
	to_fp = fdopen(to_fd, "w");
	if (to_fp == NULL)
	    r = 1;
	current_language = -1;
    }
    if (r == 0)
	fp_ref++;
    else
	return NULL;

    if (strcasecmp(arg, "german") == 0) {
	german_state.param[S_SPEED]  = 100;
	german_state.param[S_PITCH]  = 100;
	german_state.param[S_VOLUME] =  75;
	german_state.initialized = 1;
	
	s = &german_softspeech;
    } else if (strcasecmp(arg, "english") == 0) {
	english_state.param[S_SPEED]  = 100;
	english_state.param[S_PITCH]  = 100;
	english_state.param[S_VOLUME] =  75;
	english_state.initialized = 1;
	
	s = &english_softspeech;
    } else {
	s = NULL;
    }

    return s;
}


/*
 * ----------------------------------------------------------------------
 * General close. Decrement refcount, do real close when count reaches
 * zero.
 * ----------------------------------------------------------------------
 */
static int s_close(synth_t *s)
{
    int   result = 0;
    int   status;
    pid_t pid;

    assert(s->state->initialized);
    assert(fp_ref > 0);
    s->state->initialized = 0;

    fp_ref--;
    if (fp_ref == 0) {
	fclose(to_fp);
	to_fp   = NULL;
	pid = waitpid(speech_pid, &status, 0);
	if (WIFEXITED(status))
	    result = WEXITSTATUS(status);
	else if (WIFSIGNALED(status))
	    result = WTERMSIG(status);
	else
	    result = 1;
    }

    return result;
}


/*
 * ----------------------------------------------------------------------
 * Verify that the synthesizer is set to the correct language.
 * Switch if necessary and reset parameters.
 * It is necessary to give a short break to Softspeech. Somewhere 
 * there is a timing bug left.
 * ----------------------------------------------------------------------
 */
static void verify_language(struct synth_struct *s)
{
    int p;

    if (s->lang->lang == LANG_GERMAN &&
	current_language != LANG_GERMAN) {
	fprintf(to_fp, "@=2\n");
	fflush(to_fp);
	usleep(500000);		/* wait 0.5 s */
	current_language = LANG_GERMAN;
	for (p = 0; p < S_MAX; p++) {
	    s_set_param(s, p, s->state->param[p]);
	}
	sync_mark_no = 0;
	usleep(500000);		/* wait 0.5 s */
	gettimeofday(&last_command_time, NULL);
    }
    if (s->lang->lang == LANG_BRITISH_ENGLISH &&
	current_language != LANG_BRITISH_ENGLISH) {
	fprintf(to_fp, "@=1\n");
	fflush(to_fp);
	usleep(500000);		/* wait 0.5 s */
	current_language = LANG_BRITISH_ENGLISH;
	for (p = 0; p < S_MAX; p++) {
	    s_set_param(s, p, s->state->param[p]);
	}
	sync_mark_no = 0;
	usleep(500000);		/* wait 0.5 s */
	gettimeofday(&last_command_time, NULL);
    }
}


/*
 * ----------------------------------------------------------------------
 * Milliseconds from t1 (start) to t2 (end).
 * ----------------------------------------------------------------------
 */
static int elapsed_msec(struct timeval *t1, struct timeval *t2)
{
    unsigned diff;

    diff = ((t2->tv_sec - t1->tv_sec) * 1000L +
            (t2->tv_usec - t1->tv_usec) / 1000L);

    return diff;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int s_synth(struct synth_struct *s, unsigned char *buffer)
{
    assert(s->state->initialized);

    verify_language(s);
    fwrite(buffer, 1, strlen(buffer), to_fp);
    fflush(to_fp);

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int s_flush(synth_t *s)
{
    assert(s->state->initialized);

    fputc('\n', to_fp);
    fflush(to_fp);

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
static int s_clear(synth_t *s)
{
    struct timeval current_time;
    long           elapsed;

    assert(s->state->initialized);

    gettimeofday(&current_time, NULL);
    elapsed = elapsed_msec(&last_command_time, &current_time);
    if (elapsed < 1000)
	usleep(1000 * (1000 - elapsed));
    verify_language(s);

    fputc(0x18, to_fp);
    fflush(to_fp);

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * A sync mark is set by sending a string like "@I1\n" (other numbers
 * possible). When it is reached, a line "@I1\n" is sent back.
 * Return 1 when we have reached the mark, 0 when not, 2 on timeout,
 * and -1 on error.
 * ----------------------------------------------------------------------
 */
static int s_sync(synth_t *s, sync_par_t par, int timeout)
{
    struct timeval interval, cur;
    fd_set         readfs, exfs;
    char           line[100];
    int            l;
    int            flags;
    int            res = 0;
    int            gone;
    int            idx;		/* received mark number */

    assert(s->state->initialized);

    verify_language(s);

    switch (par) {
    case SYNC_POLL:		/* have we reached the mark? */
	flags = fcntl(from_fd, F_GETFL);
	if (flags < 0) {
	    perror("fcntl on pipe");
	    return -1;
	}
	if (fcntl(from_fd, F_SETFL, flags | O_NONBLOCK) < 0) {
	    perror("fcntl on pipe");
	    return -1;
	}
	l = read(from_fd, line, 100);
	if (l < 0) {
	    perror("read for sync failed");
	    return -1;
	}
	line[l] = 0;		/* end of string marker */
	idx = 0;
	sscanf(line, "@I%d\n", &idx);
	/*
	 * The index is valid under the following conditions:
	 *  - we have really read somethin (l > 0)
	 *  - it was a valid number (idx > 0)
	 *  - It is the index we expect.
	 */
	if (l > 0 && idx != 0 && idx == sync_mark_no + 1) {
	    res = 1;
	} else {
	    res = 0;
	}
	if (fcntl(from_fd, F_SETFL, flags) < 0) {
	    perror("fcntl on pipe");
	    return -1;
	}
	/*
	 * In case we did not get the right index, check for timeout.
	 */
	if (res == 0) {
	    gettimeofday(&cur, NULL);
	    if (elapsed_msec(&mark, &cur) > timeout)
		res = 2;
	}
	break;
    case SYNC_SET:		/* set mark */
	sync_mark_no = (sync_mark_no + 1) % 100;
	fprintf(to_fp, "@I%d\n", sync_mark_no + 1);
	fflush(to_fp);
	gettimeofday(&mark, NULL);
	res = 0;
	break;
    case SYNC_SET_WAIT:	/* set mark and wait (combined) */
	sync_mark_no = (sync_mark_no + 1) % 100;
	fprintf(to_fp, "@I%d\n", sync_mark_no + 1);
	fflush(to_fp);
	gettimeofday(&mark, NULL);
	/* no break, fall through */
    case SYNC_WAIT:		/* wait for mark */
	do {
	    gettimeofday(&cur, NULL);
	    gone = elapsed_msec(&mark, &cur);
	    if (gone < timeout) {
		/*
		 * Compute time that is left over and use this a timeout
		 * in select.
		 */
		interval.tv_sec  = (timeout - gone) / 1000;
		interval.tv_usec = 1000 * ((timeout - gone) % 1000);
		FD_ZERO(&readfs);
		FD_ZERO(&exfs);
		FD_SET(from_fd, &readfs);
		FD_SET(from_fd, &exfs);
		l = select(from_fd + 1, &readfs, NULL, &exfs, &interval);
		if (l < 0) {
		    perror("select failed");
		    return -1;
		}
		if (FD_ISSET(from_fd, &exfs)) {
		    fprintf(stderr, "expection on fd\n");
		    return -1;
		}
		if (FD_ISSET(from_fd, &readfs)) {
		    l = read(from_fd, line, 100);
		    if (l < 0) {
			perror("read for sync failed");
			return -1;
		    }
		    line[l] = 0;		/* end of string marker */
		    idx = 0;
		    sscanf(line, "@I%d\n", &idx);
		    /*
		     * The index is valid under the following conditions:
		     *  - we have really read somethin (l > 0)
		     *  - it was a valid number (idx > 0)
		     *  - It is the index we expect.
		     */
		    if (l > 0 && idx != 0 && idx == sync_mark_no + 1) {
			res = 1;
		    } else {
			res = 0;
		    }
		}
	    } else {
		res = 2;	/* timeout */
	    }
	} while (res == 0);
	break;
    default:			/* should not happen */
	res = -1;
	break;
    }

    return res;
}


/*
 * ----------------------------------------------------------------------
 * Get a synthesizer parameter.
 * ----------------------------------------------------------------------
 */
static int s_get_param(struct synth_struct *s, synth_par_t par, int *value)
{
    if (par >= 0 && par < S_MAX) {
	*value = s->state->param[par];
	return 0;
    } else
	return 1;
}


/*
 * ----------------------------------------------------------------------
 * Set a parameter of the synthesizer. There is one problem: When we
 * clear immediately after issuing a command, the command may be lost.
 * For this reason, record the time of the command and sleep in the
 * clear function when there was a command in the last second.
 * Setting is only forwarded to the synthesizer if the selected 
 * synthesizer is the currently active one.
 * ----------------------------------------------------------------------
 */
static int s_set_param(struct synth_struct *s, synth_par_t par, int value)
{

    switch (par) {
    case S_SPEED:
	if (value <   10) value =   10;
	if (value > 1000) value = 1000;
	if (s->lang->lang == current_language) {
	    gettimeofday(&last_command_time, NULL);
	    fprintf(to_fp, "@W%d\n", value);
	    fflush(to_fp);
	}
	s->state->param[par] = value;
	break;
    case S_PITCH:
	if (value <   10) value =   10;
	if (value > 1000) value = 1000;
	if (s->lang->lang == current_language) {
	    gettimeofday(&last_command_time, NULL);
	    fprintf(to_fp, "@F%d\n", value);
	    fflush(to_fp);
	}
	s->state->param[par] = value;
	break;
    case S_VOLUME:
	if (value <   0) value =   0;
	if (value > 100) value = 100;
	if (s->lang->lang == current_language) {
	    gettimeofday(&last_command_time, NULL);
	    fprintf(to_fp, "@A%d\n", value);
	    fflush(to_fp);
	}
	s->state->param[par] = value;
	break;
    default:
	/* ignore */
    }

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * Start a program (int the background) and connect it with two pipes
 * to the caller:
 *  *to from caller to program
 *  *from from programm to caller
 * Process id of the program is returned in *pid.
 * ----------------------------------------------------------------------
 */
static int start_program(char *argv[], int *to, int *from, int *pid)
{
    int to_pipe[2], from_pipe[2];
    int i;

    if (pipe(to_pipe) < 0 || pipe(from_pipe) < 0) {
	perror("create pipes");
	return -1;
    }
    *pid = fork();
    assert(*pid >= 0);
    if (*pid == 0) {		/* child */
	close(from_pipe[0]);
	close(to_pipe[1]);
	if (dup2(to_pipe[0], 0) < 0 || dup2(from_pipe[1], 1) < 0) {
	    perror("dup2 on pipes failed");
	    kill(getppid(), SIGKILL);
	    exit(1);
	}
	close(from_pipe[1]);
	close(to_pipe[0]);
	if (chdir(SPEECH_DIR) < 0) {
	    perror("change directory");
	    kill(getppid(), SIGKILL);
	    exit(1);
	}
	execvp(argv[0], argv);	/* should never return */
	perror("exec failed");
	fprintf(stderr, "command and args:");
	for (i = 0; argv[i] != NULL; i++)
	    fprintf(stderr, " %s ", argv[i]);
	fprintf(stderr, "\n");
	kill(getppid(), SIGKILL);
	exit(1);
    } else {			/* parent */
	*to   = to_pipe[1];
	*from = from_pipe[0];
	close(to_pipe[0]);
	close(from_pipe[1]);
    }

    return 0;
}
