/*
 * ======================================================================
 * Test program for speech library.
 * Copyright (C) 1998-1999 Roger Butenuth <butenuth@uni-paderborn.de>
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 * For the latest version see:
 * http://www.uni-paderborn.de/cs/heiss/blinux/index.html
 * $Id: speech_test.c,v 1.11 1999/03/11 21:08:59 butenuth Exp $
 * ======================================================================
 */
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include "language.h"

/*
 * Prototypes for the open functions.
 */
lang_open_f german_open, english_open;

/*
 * Handles for active languages.
 */
lang_t *lang_eng, *lang_ger;


int main(int argc, char *argv[])
{
    int i, value, s;

    /*
     * Calling open with NULL as parameter creates a dummy language module.
     * All functions are still there, but no output is done. 
     */
    lang_ger = german_open(NULL);
    assert(lang_ger != NULL);
    /*
     * Here we open a real synthesizer. The first part of the string (up
     * to the first blank character) is the name of a shared library. 
     * This must be an absolute file name or the library has to be in
     * one of the library directories. 
     * Everything after the first blank is given to the synthesizer and
     * specific to that synthesizer. "softspeech" supports german and 
     * english, so you must specify it here.
     */
    lang_eng = english_open("/home/butenuth/softspeech.so english");
    assert(lang_eng != NULL);
    /*
     * A language and a synthesizer have configurable parameters. Read
     * and print the language parameters.
     */
    printf("parameters for german language:\n");
    for (i = 0; i < L_MAX; i++) {
	lang_ger->get_param(lang_ger, i, &value);
	printf("param[%d] = %d\n", i, value);
    }

    printf("parameters for english language:\n");
    for (i = 0; i < L_MAX; i++) {
	lang_eng->get_param(lang_eng, i, &value);
	printf("param[%d] = %d\n", i, value);
    }
    /*
     * Speak some german text. With the dummy synthesizer loaded, this
     * will produce nothing but silence...
     */
    printf("german text:\n");
    lang_ger->speak_string(lang_ger,
			   "Fischers Fritz fischt frische Fische.\n");
    /*
     * SYNC_SET_WAIT means: Set a mark in the speech, wait until that
     * mark has been reached, but do not wait longer than 2000 ms.
     * Result is:
     *  < 0: error
     *    0: mark not reached (for polling)
     *    1: mark reached, no timeout
     *    2: timeout
     * With a real synth and decent speed you should expect a timeout
     * in this case.
     */
    s = lang_ger->synth->sync(lang_ger->synth, SYNC_SET_WAIT, 2000);
    printf("sync done, s = %d\n", s);
    lang_ger->speak_string(lang_ger,
			   "Frische Fische fischt Fischers Fritz.\n");
    /*
     * Another mark, 10 s should be enough, so expect no timeout.
     */
    lang_ger->synth->sync(lang_ger->synth, SYNC_SET_WAIT, 10000);
    /*
     * Now speek an english sentence.
     */
    lang_eng->speak_string(lang_eng,
			   "To be or not to be thats the question\n");
    /*
     * Give the synthesizer one second time, after that, interrupt
     * it and remove everything in the speech queue.
     */
    sleep(1);
    lang_eng->synth->clear(lang_eng->synth);
    /*
     * Speak a second sentence and wait until it has been spoken.
     */
    lang_eng->speak_string(lang_eng,
			   "This is the second sentence.\n");
    s = lang_eng->synth->sync(lang_eng->synth, SYNC_SET_WAIT, 10000);
    printf("last sync done, s = %d, close\n", s);
    /*
     * Close both languages/synthesizers.
     */
    lang_ger->close(lang_ger);
    lang_eng->close(lang_eng);

    return 0;
}
