/*
 * BRLTTY - Access software for Unix for a blind person
 *          using a soft Braille terminal
 *
 * Copyright (C) 1995-1998 by The BRLTTY Team, All rights reserved.
 *
 *     Nicolas Pitre <nico@cam.org>
 *     Stphane Doyon <s.doyon@videotron.ca>
 *     Nikhil Nair <nn201@cus.cam.ac.uk>
 * This part changed by:
 *     Roger Butenuth <butenuth@uni-paderborn.de>
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 *
 * $Id: brltty.c,v 1.26 1999/03/11 21:08:28 butenuth Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <dlfcn.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <errno.h>

#include "config.h"
#include "config_file.h"
#include "setup.h"
#include "brl.h"
#include "scr.h"
#include "misc.h"
#include "Speech/language.h"
#include "Speech/synthesizer.h"

#define VERSION "BRLTTY experimental version"



#define USAGE "\
Usage: %s [options] \n\
 -c config-file       use configuration file `config-file' \n\
 -h, --help           print this usage message \n\
 -q, --quiet          suppress start-up messages \n\
 -l n                 debugging level for syslog (from 0 to 7, default 4) \n\
 -v, --version        print start-up messages and exit \n"


/*
 * Some useful macros: 
 */
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define BRL_ISUPPER(c) (isupper (c) || (c) == '@' || (c) == '[' || (c) == '^' \
			|| (c) == ']' || (c) == '\\')

#define TOGGLE(var, keypress) \
    (var = ((keypress) & VAL_SWITCHON) ? 1 \
     : ((keypress) & VAL_SWITCHOFF) ? 0 \
     : var ^ 1)

/*
 * Argument for usetable() 
 */
#define TBL_TEXT   0		/* use text translation table */
#define TBL_ATTRIB 1		/* use attribute translation table */


typedef struct {
    short csron;		/* display cursor on (toggled during blink) */
    short csrcntr;
    short capon;		/* display caps off (toggled during blink) */
    short capcntr;
    short attron;		/* display attritubes off (toggled d. b.) */
    short attrcntr;
} blink_state_t;

int  initialize(int argc, char *argv[]);
int  finalize(void);
int  deamonize(void);
void close_speech(void);
void main_loop(void);
int  handle_input_event(int keypress, blink_state_t *blink, scrstat_t *scr);
void do_display(scrstat_t *scr, blink_state_t *blink);
void init_braille(brl_drv_t *brl_drv);
void init_speech(speech_drv_t *speech_drv);


int s_csrvis        = INIT_CSRVIS;
int s_attrvis       = INIT_ATTRVIS;
int s_csrblink      = INIT_CSRBLINK;
int s_capblink      = INIT_CAPBLINK;
int s_attrblink     = INIT_ATTRBLINK;
int s_csrsize       = INIT_CSRSIZE;
int s_csroncnt      = INIT_CSR_ON_CNT;
int s_csroffcnt     = INIT_CSR_OFF_CNT;
int s_caponcnt      = INIT_CAP_ON_CNT;
int s_capoffcnt     = INIT_CAP_OFF_CNT;
int s_attroncnt     = INIT_ATTR_ON_CNT;
int s_attroffcnt    = INIT_ATTR_OFF_CNT;
int s_sixdots       = INIT_SIXDOTS;
int s_slidewin      = INIT_SLIDEWIN;

int s_speech        = INIT_SPEECHON;
int s_skip_id_lines = INIT_SKPIDLNS;


int           no_speech_chans;
speech_chan_t *speech_chan;	/* array [0..no_speech_chans - 1] */

brltty_param_t initparam = {
    INIT_CSRTRK, TBL_TEXT, 0, 0, 0, 0, NULL
};

/*
 * Array definition containing brltty_param structures for 
 * each screen.
 * Note: Linux counts consoles from 1, brltty from 0,
 *       so we waste number zero. Don't mind.
 */
struct brltty_param scrparam[NBR_SCR];
struct brltty_param *p;		/* pointer to current param structure */
int                 curscr;	/* current screen number */

/* Misc param variables */
int dispmd = LIVE_SCRN;	/* freeze screen on/off */
int infmode = 0;		/* display screen image or info */
int csr_offright;		/* used for sliding window */
int hwinshift;		/* Half window horizontal distance */
int fwinshift;		/* Full window horizontal distance */
int vwinshift;		/* Window vertical distance */

brl_disp_t brl;		/* For the Braille routines */


/*
 * Output translation tables - the files *.auto.h are generated at *
 * compile-time: 
 */
unsigned char texttrans[256] =
{
#include "text.auto.h"
};

unsigned char attribtrans[256] =
{
#include "attrib.auto.h"
};

unsigned char *curtbl = NULL;	/* currently active translation table */

volatile sig_atomic_t keep_going = 1;	/* controls program termination */
int                   tbl_fd;	/* Translation table filedescriptor */

/* -h, -l, -q and -v options */
int opt_h = 0, opt_q = 0, opt_v = 0, opt_l = 4;
char *conf_file  = "/etc/brltty/config"; /* human readable configuration */
char *state_file = "/etc/brltty/state";	 /* current state (generated) */
char *table_file = NULL;		 /* text translation table filename */

syntax_tree_t *conf_tree = NULL;

unsigned int TickCount = 0;	/* incremented each cycle */

/* 
 * Number dot translation for status cells 
 */
const unsigned char num[10] = { 14, 1, 5, 3, 11, 9, 7, 15, 13, 6 };

/*
 * Function prototypes: 
 */
void start_braille(brl_drv_t *brl_drv);
void switchto(unsigned int scrno); /* activate params for specified screen */
void setwinxy(scrstat_t *scr);	/* move window to current position */
void clrbrlstat(void);
void error_handler(int signum);	/* handle unexpected signals */
void termination_handler(int signum);	/* clean up before termination */
void usetable(int tbl);		/* set character translation table */
void configmenu(void);		/* configuration menu */

#if defined (Papenmeier)
int pm_num(int x);
int pm_stat(int line, int on);
#else
int pm_num(int x) { return 0; }
int pm_stat(int line, int on) { return 0; }
#endif


/*
 * ----------------------------------------------------------------------
 * Speak current line over the currently active speech channel.
 * ----------------------------------------------------------------------
 */
int say_line(int line, int len)
{
    unsigned char  buffer[len + 1];
    winpos_t       pos;
    synth_t *synth = p->speech->lang->synth;
    lang_t  *lang  = p->speech->lang;

    if (s_speech) {
	pos.left = 0;
	pos.top  = line;
	pos.width  = len;
	pos.height = 1;
	getscr(pos, buffer, SCR_TEXT);
	buffer[len] = 0;
	
	synth->clear(synth);
	lang->speak_string(lang, buffer);
    }

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * Small and simple: init, main loop, terminate.
 * ----------------------------------------------------------------------
 */
int main(int argc, char *argv[])
{
    int retval;

    retval = initialize(argc, argv);
    if (retval)
	return retval;

    main_loop();

    retval = finalize();

    return retval;
}


/*
 * ----------------------------------------------------------------------
 * Parse command line, read config file, deamonize.
 * ----------------------------------------------------------------------
 */
int initialize(int argc, char *argv[])
{
    char      *endptr;
    int       i;

    /*
     * Parse command line using getopt():
     */
    while ((i = getopt (argc, argv, "c:hl:qv-:")) != -1)
	/*
	 * This will complain if an incorrect option is given but will still
	 * proceed.
	 */
	switch (i) {
	case 'c':		/* configuration filename */
	    conf_file = optarg;
	    break;
	case 'h':		/* help */
	    opt_h = 1;
	    break;
	case 'l':		/* log level */
	    opt_l = strtol(optarg,&endptr,0);
	    if(endptr==optarg || *endptr != 0 || opt_l<0 || opt_l>7) {
		fprintf(stderr,"Invalid log level... ignored\n");
		opt_l = 4;
	    }
	    break;
	case 'q':		/* quiet */
	    opt_q = 1;
	    break;
	case 'v':		/* version */
	    opt_v = 1;
	    break;
	case '-':		/* long options */
	    if (strcmp (optarg, "help") == 0)
		opt_h = 1;
	    else if (strcmp (optarg, "quiet") == 0)
		opt_q = 1;
	    else if (strcmp (optarg, "version") == 0)
		opt_v = 1;
	    break;
	}
    
    if (opt_h) {
	printf (USAGE, argv[0]);
	return 0;
    }
    /*
     * Print version and copyright information: 
     */
    if (!(opt_q && !opt_v)) {	
	/*
	 * keep quiet only if -q and not -v
	 */
	printf("%s starting\n", VERSION);
	printf("Copyright (C) 1995-1998 by The BRLTTY Team."
	       "All rights reserved.\n"
);
    }
    if (opt_v)
	exit(0);
    /*
     * Read configuration file.
     */
    if (parse_config_file(conf_file, &conf_tree)) {
	fprintf(stderr, "Could not read configuration file %s\n", conf_file);
	 exit(1);
    }
    if (conf_tree->brl_drv == NULL) {
	fprintf(stderr, "No Braille driver specified\n");
	exit(1);
    }

    if (chdir (HOME_DIR))	/* change to directory containing data files */
	chdir ("/etc");		/* home directory not found, use backup */
    /*
     * Load text translation table: 
     */
    if (table_file) {
	if ((tbl_fd = open(table_file, O_RDONLY)) >= 0 &&
	    (curtbl = (unsigned char *) malloc(256)) &&
	    read(tbl_fd, curtbl, 256) == 256)
	    memcpy(texttrans, curtbl, 256);
	else
	    fprintf(stderr, "Failed to read dot translation table %s",
		    table_file);
	if (curtbl)
	    free(curtbl);
	if (tbl_fd >= 0)
	    close(tbl_fd);
    }
    /*
     * Initialize screen library 
     */
    if (init_screen()) {				
	fprintf(stderr, "Cannot read screen\n");
	exit(2);
    }
    /*
     * Initialise all screen structures with default values.
     * "speech" is initialized with NULL at this point, it is filled
     * with something more meaingful in init_speech().
     */
    for (i = 0; i < NBR_SCR; i++)
	scrparam[i] = initparam;
    /*
     * Let's go:
     *  - Initialize speech
     *  - Initialize setup (speech must be ready at this time)
     *  - Initialize Braille and set text display
     *  - Load setup from file, this must be the last action because
     *    it results in a lot of set-value-calls.
     */
    init_speech(conf_tree->speech_drv);
    init_setup();
    init_braille(conf_tree->brl_drv);
    load_setup(state_file);
    LogOpen(opt_l);
    deamonize();
    /*
     * Establish signal handler to clean up before termination: 
     */
    if (signal(SIGTERM, termination_handler) == SIG_IGN)
	signal(SIGTERM, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGHUP, SIG_IGN);
    signal(SIGPIPE, error_handler);
    signal(SIGCHLD, error_handler);
    signal(SIGSEGV, error_handler);

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * Clean up.
 * ----------------------------------------------------------------------
 */
int finalize(void)
{
    clrbrlstat();
    message("BRLTTY terminating");
    close_screen();
    close_speech();
    brl.closebrl(&brl);
    LogPrint(LOG_NOTICE,"Terminating");
    LogClose();
    return 0;
}


/*
 * ----------------------------------------------------------------------
 * Become a daemon: 
 * ----------------------------------------------------------------------
 */
int deamonize(void)
{
    switch (fork ()) {
    case -1:			/* can't fork */
	perror("fork()");
	close_screen();
	LogClose();
	exit(3);
    case 0:			/* child, process becomes a daemon: */
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);
	LogPrint(LOG_DEBUG, "Becoming daemon");
	/*
	 * request a new session (job control)
	 */
	if (setsid() == -1) {			
	    close_screen();
	    LogPrint(LOG_ERR, "setsid: %s", strerror(errno));
	    LogClose();
	    exit(4);
	}
	break;
    default:			/* parent returns to calling process: */
	exit(0);
    }

    return 0;
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void main_loop(void)
{
    blink_state_t blink;	/* for blinking caps, cursor, and attributes */
    scrstat_t     scr;		/* For screen statistics */
    int           keypress;	/* character received from braille display */
    int           oldwinx, oldwiny;
    int           say;

    blink.csron    = 1;
    blink.csrcntr  = 1;
    blink.capon    = 0;
    blink.capcntr  = 1;
    blink.attron   = 0;
    blink.attrcntr = 1;
    /*
     * Get current active console and activate it. This must be done
     * early because switching to a screen allocates memory.
     */
    getstat(&scr);
    switchto(scr.no);
    setwinxy(&scr);

    oldwinx = p->winx; oldwiny = p->winy;
    p->cox = scr.posx;
    p->coy = scr.posy;
    /*
     * Main program loop 
     */
    while (keep_going) {
	TickCount++;
	say = 0;
	/*
	 * Process any Braille and keyboard input.
	 * When there is nor more input available, start to update display.
	 */
	while ((keypress = get_input_event(TBL_CMD, EVNT_NON_BLOCKING)) !=
	       EOF) {
	    say |= handle_input_event(keypress, &blink, &scr);
	}
	/*
	 * Update blink counters: 
	 */
	if (s_csrblink && !--blink.csrcntr)
	    blink.csrcntr = (blink.csron ^= 1) ? s_csroncnt : s_csroffcnt;
	if (s_capblink &&!--blink.capcntr)
	    blink.capcntr = (blink.capon ^= 1) ? s_caponcnt : s_capoffcnt;
	if (s_attrblink && !--blink.attrcntr)
	    blink.attrcntr = (blink.attron ^= 1)
		? s_attroncnt : s_attroffcnt;
	/*
	 * Update Braille display and screen information.
	 * Switch screen params if screen number has changed.
	 */
	getstat(&scr);
	if (!(dispmd & FROZ_SCRN) && curscr != scr.no)
	    switchto(scr.no);
	/*
	 * cursor tracking
	 */
	if (p->csrtrk) {
	    /*
	     * If cursor moves while blinking is on
	     */
	    if (s_csrblink) {
		if (scr.posy != p->coy) {
		    /*
		     * turn off cursor to see what's under it while changing
		     * lines
		     */
		    blink.csron = 0;
		    blink.csrcntr = s_csroffcnt;
		}
		if (scr.posx != p->cox) {
		    /*
		     * turn on cursor to see it moving on the line
		     */
		    blink.csron = 1;
		    blink.csrcntr = s_csroncnt;
		}
	    }
	    /*
	     * If the cursor moves in cursor tracking mode:
	     */
	    if (scr.posx != p->cox || scr.posy != p->coy) {
		setwinxy(&scr);
		if (scr.posy != p->coy)
		    say = 1;
		p->cox = scr.posx;
		p->coy = scr.posy;
	    }
	}
	/*
	 * If attribute underlining is blinking during display movement
	 */
	if (s_attrvis && s_attrblink) {
	    /*
	     * We could check to see if we changed screen, but that doesn't
	     * really matter... this is mainly for when you are hunting up/down
	     * for the line with attributes.
	     */
	    if(p->winx != oldwinx || p->winy != oldwiny) {
		blink.attron = 1;
		blink.attrcntr = s_attroncnt;
	    }
	    /*
	     * problem: this still doesn't help when the braille window is
	     * stationnary and the attributes themselves are moving
	     * (example: tin).
	     */
	}
	oldwinx = p->winx; oldwiny = p->winy;
	/*
	 * Set status cells and main part of display
	 */
	do_display(&scr, &blink);
	if (say)
	    say_line(p->winy, scr.cols);
	delay (DELAY_TIME);
    }
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
int handle_input_event(int keypress, blink_state_t *blink, scrstat_t *scr)
{
    int say = 0;

    switch (keypress & ~VAL_SWITCHMASK) {
    case CMD_RESTARTBRL:
	/* @@@ to be done */
	break;
    case CMD_TOP:
	p->winy = 0;
	break;
    case CMD_TOP_LEFT:
	p->winy = 0;
	p->winx = 0;
	break;
    case CMD_BOT:
	p->winy = scr->rows - brl.y;
	break;
    case CMD_BOT_LEFT:
	p->winy = scr->rows - brl.y;
	p->winx = 0;
	break;
    case CMD_WINUP:
	p->winy = MAX (p->winy - vwinshift, 0);
	say = 1;
	break;
    case CMD_WINDN:
	p->winy = MIN (p->winy + vwinshift, scr->rows - brl.y);
	say = 1;
	break;
    case CMD_LNUP:
	say = 1;
	if (p->winy == 0) {
	    break;
	}
	if (!s_skip_id_lines) {
	    p->winy = MAX(p->winy - 1, 0);
	    break;
	}
	/* no break here */
    case CMD_PRDIFLN:
    case CMD_ATTRUP:
	if (p->winy != 0) {
	    char buffer1[scr->cols], buffer2[scr->cols];
	    int scrtype = (p->dispmode || keypress==CMD_ATTRUP)
		? SCR_ATTRIB : SCR_TEXT;
	    getscr ((winpos_t) {
		0, p->winy, scr->cols, 1
		    }
		,buffer1, scrtype);
	    while (p->winy > 0) {
		getscr ((winpos_t)
			{ 0, --p->winy, scr->cols, 1 },
			buffer2, scrtype);
		if (memcmp (buffer1, buffer2, scr->cols) ||
		    (keypress != CMD_ATTRUP && p->winy == scr->posy))
		    break;	/* lines are different */
		if(p->winy == 0) {
		    break;
		}
	    }
	}
	break;
    case CMD_LNDN:
	say = 1;
	if (p->winy == scr->rows - brl.y)
	    break;
	if (!s_skip_id_lines) {
	    p->winy = MIN(p->winy + 1, scr->rows - brl.y);
	    break;
	}
	/* no break here */
    case CMD_NXDIFLN:
    case CMD_ATTRDN:
	if (p->winy != scr->rows - brl.y) {
	    char buffer1[scr->cols], buffer2[scr->cols];
	    int scrtype = (p->dispmode || keypress==CMD_ATTRDN)
		? SCR_ATTRIB : SCR_TEXT;
	    getscr ((winpos_t) { 0, p->winy, scr->cols, 1 }, buffer1, scrtype);
	    while (p->winy < scr->rows - brl.y) {
		getscr ((winpos_t) { 0, ++p->winy, scr->cols, 1 },
		        buffer2, scrtype);
		if (memcmp (buffer1, buffer2, scr->cols) || 
		    (keypress!=CMD_ATTRDN && p->winy == scr->posy))
		    break;	/* lines are different */
		if(p->winy == scr->rows-brl.y) {
		    break;
		}
	    }
	}
	break;
    case CMD_HOME:
	setwinxy(scr);
	say = 1;
	break;
    case CMD_LNBEG:
	p->winx = 0;
	break;
    case CMD_LNEND:
	p->winx = scr->cols - brl.x;
	break;
    case CMD_CHRLT:
	p->winx = MAX(p->winx - 1, 0);
	break;
    case CMD_CHRRT:
	p->winx = MIN(p->winx + 1, scr->cols - brl.x);
	break;
    case CMD_HWINLT:
	p->winx = MAX(p->winx - hwinshift, 0);
	break;
    case CMD_HWINRT:
	p->winx = MIN(p->winx + hwinshift, scr->cols - brl.x);
	break;
    case CMD_FWINLT:
	if (p->winx == 0 && p->winy > 0) {
	    p->winx = scr->cols - brl.x;
	    p->winy--;
	} else if (!(p->winx == 0 && p->winy == 0))
	    p->winx = MAX(p->winx - fwinshift, 0);
	break;
    case CMD_FWINRT:
	if (p->winx == scr->cols - brl.x && p->winy < scr->rows - brl.y) {
	    p->winx = 0;
	    p->winy++;
	} else if (!(p->winx == scr->cols - brl.x &&
		     p->winy == scr->rows - brl.y))
	    p->winx = MIN(p->winx + fwinshift, scr->cols - brl.x);
	break;
    case CMD_CSRJMP:
	cursor_routing(curscr, p->winx, p->winy);
	break;
    case CMD_CSRJMP_VERT:
	cursor_routing(curscr, -1, p->winy);
	break;
#if 0
    case CMD_KEY_UP:
	inskey(UP_CSR);
	break;
    case CMD_KEY_DOWN:
	inskey(DN_CSR);
	break;
    case CMD_KEY_RIGHT:
	inskey(RT_CSR);
	break;
    case CMD_KEY_LEFT:
	inskey(LT_CSR);
	break;
    case CMD_KEY_RETURN:
	inskey(KEY_RETURN);
	break;
#endif
    case CMD_CSRVIS:
	TOGGLE(s_csrvis, keypress);
	break;
    case CMD_ATTRVIS:
	TOGGLE(s_attrvis, keypress);
	break;
    case CMD_CSRBLINK:
	blink->csron = 1;
	TOGGLE(s_csrblink, keypress);
	if (s_csrblink) {
	    blink->csrcntr = s_csroncnt;
	    blink->attron = 1;
	    blink->attrcntr = s_attroncnt;
	    blink->capon = 0;
	    blink->capcntr = s_capoffcnt;
	}
	break;
    case CMD_CSRTRK:
	TOGGLE(p->csrtrk, keypress);
	if (p->csrtrk)
	    setwinxy(scr);
	break;
    case CMD_CUT_BEG:
	cut_begin(p->winx, p->winy);
	break;
    case CMD_CUT_END:
	cut_end(p->winx + brl.x - 1, p->winy + brl.y - 1);
	break;
    case CMD_PASTE:
	cut_paste ();
	break;
    case CMD_SND:
	; /* no longer used (interferes with speech) */
	break;
    case CMD_DISPMD:
	usetable((TOGGLE(p->dispmode, keypress)) ? TBL_ATTRIB : TBL_TEXT);
	break;
    case CMD_FREEZE:
	dispmd = selectdisp(dispmd ^ FROZ_SCRN);
	break;
#if 0
    case CMD_HELP:
	infmode = 0;	/* disable info mode */
	dispmd = selectdisp(dispmd ^ HELP_SCRN);
	break;
#endif
    case CMD_CAPBLINK:
	blink->capon = 1;
	TOGGLE(s_capblink, keypress);
	if (s_capblink) {
	    blink->capcntr = s_caponcnt;
	    blink->attron = 0;
	    blink->attrcntr = s_attroffcnt;
	    blink->csron = 0;
	    blink->csrcntr = s_csroffcnt;
	}
	break;
    case CMD_ATTRBLINK:
	blink->attron = 1;
	TOGGLE(s_attrblink, keypress);
	if (s_attrblink) {
	    blink->attrcntr = s_attroncnt;
	    blink->capon = 1;
	    blink->capcntr = s_caponcnt;
	    blink->csron = 0;
	    blink->csrcntr = s_csroffcnt;
	}
	break;
    case CMD_INFO:
	TOGGLE(infmode, keypress);
	break;
    case CMD_CSRSIZE:
	TOGGLE(s_csrsize, keypress);
	break;
    case CMD_SIXDOTS:
	TOGGLE(s_sixdots, keypress);
	break;
    case CMD_SLIDEWIN:
	TOGGLE(s_slidewin, keypress);
	break;
    case CMD_SKPIDLNS:
	TOGGLE(s_skip_id_lines, keypress);
	break;
#if 0
    case CMD_SAVECONF:
	saveconfig();
	break;
#endif
    case CMD_CONFMENU:
	setup(curscr, scrparam);
	save_setup(state_file);
	break;
#if 0
    case CMD_RESET:
	loadconfig();
#endif
	blink->csron = 1;
	blink->capon = 0;
	blink->csrcntr = 1;
	blink->capcntr = 1;
	break;
    case CMD_SAY:
	say = 1;
	break;
#if 0
    case CMD_MUTE:
	mutespk ();
	break;
#endif
    default:
	if (keypress >= CR_ROUTEOFFSET && keypress < CR_ROUTEOFFSET + brl.x) {
	    /*
	     * Cursor routing keys:
	     */
	    cursor_routing(curscr,
			   p->winx + keypress - CR_ROUTEOFFSET,
			   p->winy);
	} else if (keypress >= CR_BEGBLKOFFSET &&
		   keypress < CR_BEGBLKOFFSET + brl.x)
	    cut_begin (p->winx + keypress - CR_BEGBLKOFFSET, p->winy);
	else if (keypress >= CR_ENDBLKOFFSET &&
		 keypress < CR_ENDBLKOFFSET + brl.x)
	    cut_end (p->winx + keypress - CR_ENDBLKOFFSET, p->winy);
	else
	    LogPrint(LOG_DEBUG, "Driver sent unrecognized command 0x%x\n",
		     keypress);
	break;
    }

    return say;
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void do_display(scrstat_t *scr, blink_state_t *blink)
{
    int           i;		/* loop counter */

    if (infmode) {
	/*
	 * If in info mode, send status information: 
	 */
	char infbuf[20];
	sprintf(infbuf, "%02d:%02d %02d:%02d %02d %c%c%c%c%c%c",
		p->winx, p->winy, scr->posx, scr->posy, curscr, 
		p->csrtrk ? 't' : ' ',
		s_csrvis ? (s_csrblink ? 'B' : 'v') :
		(s_csrblink ? 'b' : ' '), p->dispmode ? 'a' : 't',
		(dispmd & FROZ_SCRN) == FROZ_SCRN ? 'f' : ' ',
		s_sixdots ? '6' : '8', s_capblink ? 'B' : ' ');
	/*
	 * status cells 
	 */
	memset(brl.stat, 0, brl.s);
	brl.stat[0] = texttrans['I'];
	brl.stat[1] = texttrans['n'];
	brl.stat[2] = texttrans['f'];
	brl.stat[3] = texttrans['o'];
	brl.stat[4] = texttrans[' '];
	brl.setbrlstat(&brl);
	
	message(infbuf);
    } else {
	/*
	 * If not in info mode, get screen image and update the display.
	 * For the sake of displays on which the status cells and the
	 * main display have to be updated together, it is guaranteed
	 * that setbrlstat() will be called just *before* writebrl(). 
	 */
	switch (brl.stat_mode) {
	case ST_AlvaStyle:
	    /*
	     * The coords are given with letters as the DOS tsr 
	     */
	    brl.stat[0] = ((TickCount / 16) % (scr->posy / 25 + 1))
		? 0
		: texttrans[scr->posy % 25 + 'a'] | (scr->posx / brl.x) << 6;
	    brl.stat[1] = ((TickCount / 16) % (p->winy / 25 + 1))
		? 0
		: texttrans[p->winy % 25 + 'a'] | (p->winx / brl.x) << 6;
	    brl.stat[2] = texttrans[(p->dispmode)
				    ? 'a'
				    : ((dispmd & FROZ_SCRN) == FROZ_SCRN) ? 'f' :
				    (p->csrtrk) ? 't' : ' '];
	    brl.stat[3] = 0;
	    brl.stat[4] = 0;
	    break;
	case ST_TiemanStyle:
	    brl.stat[0] = num[(p->winx / 10) % 10] << 4 | 
		num[(scr->posx / 10) % 10];
	    brl.stat[1] = num[p->winx % 10] << 4 | num[scr->posx % 10];
	    brl.stat[2] = num[(p->winy / 10) % 10] << 4 | 
		num[(scr->posy / 10) % 10];
	    brl.stat[3] = num[p->winy % 10] << 4 | num[scr->posy % 10];
	    brl.stat[4] = s_csrvis << 1 | s_csrsize << 3 |
		s_csrblink << 5 | s_slidewin << 7 | p->csrtrk << 6 |
		s_speech << 4 | p->dispmode << 2;
	    brl.stat[4] |= (dispmd & FROZ_SCRN) == FROZ_SCRN ? 1 : 0;
	    break;
	case ST_PB80Style:
	    brl.stat[0] = num[(p->winy+1) % 10] << 4 | \
		num[((p->winy+1) / 10) % 10];
	    break;
	case ST_Papenmeier:
	    brl.stat[0] = pm_num(p->winy+1);
	    brl.stat[1] = 0;  /* empty for easier reading */
	    brl.stat[2] =  pm_num(scr->posy+1);
	    
	    brl.stat[3] =  pm_num(scr->posx+1);
	    brl.stat[4] = 0;     /* empty for easier reading */
	    brl.stat[5] = pm_stat(6, p->csrtrk);
	    brl.stat[6] = pm_stat(7, p->dispmode);
	    brl.stat[7] = 0;     /* empty for easier reading */
	    brl.stat[8] = pm_stat(9, (dispmd & FROZ_SCRN) == FROZ_SCRN);
	    brl.stat[9] = 0;
	    brl.stat[10] = 0;
	    brl.stat[11] = 0;
	    brl.stat[12] = pm_stat(3, s_csrvis);
	    brl.stat[13] = pm_stat(4, s_csrsize);
	    brl.stat[14] = pm_stat(5, s_csrblink);
	    brl.stat[15] = pm_stat(6, s_capblink);
	    brl.stat[16] = pm_stat(7, s_sixdots);
	    brl.stat[17] = pm_stat(8, s_speech);
	    brl.stat[18] = pm_stat(9, s_skip_id_lines);
	    brl.stat[19] = pm_stat(0, s_attrvis);
	    brl.stat[20] = pm_stat(1, s_attrblink);
	    brl.stat[21] = 0;
	    break;
	case ST_TSI:
	default:
	    memset(brl.stat, 0, brl.s);
	    break;
	}
	brl.setbrlstat(&brl);
	
	getscr((winpos_t) { p->winx, p->winy, brl.x, brl.y },
	brl.disp,
	p->dispmode ? SCR_ATTRIB : SCR_TEXT);
	if (s_capblink && !blink->capon)
	    for (i = 0; i < brl.x * brl.y; i++)
		if (BRL_ISUPPER (brl.disp[i]))
		    brl.disp[i] = ' ';
	/*
	 * Do Braille translation using current table: 
	 */
	if (s_sixdots && curtbl != attribtrans)
	    for (i = 0; i < brl.x * brl.y; brl.disp[i] =
		     curtbl[brl.disp[i]] & 0x3f, i++);
	else
	    for (i = 0; i < brl.x * brl.y; brl.disp[i] =
		     curtbl[brl.disp[i]], i++);
	/*
	 * Attribute underlining: if viewing text (not attributes),
	 * attribute underlining is active and visible,
	 * then we get the attributes for the current region
	 * and OR the underline.
	 */
	if (!p->dispmode && s_attrvis &&
	    (!s_attrblink || blink->attron)) {
	    unsigned char attrbuf[brl.x * brl.y];
	    getscr((winpos_t) { p->winx, p->winy, brl.x, brl.y },
	    attrbuf, 
	    SCR_ATTRIB);
	    for (i = 0; i < brl.x * brl.y; i++)
		switch (attrbuf[i]) {
		    /*
		     * Experimental! Attribute values are hardcoded...
		     */
		case 0x07: 
		case 0x17: 
		case 0x30: 
		    break;
		case 0x70:
		    brl.disp[i] |= ATTR1CHAR;
		    break;
		case 0x0F:
		default:
		    brl.disp[i] |= ATTR2CHAR;
		    break;
		}
	}
	/*
	 * If the cursor is visible and in range:
	 */
	if (s_csrvis && (!s_csrblink || blink->csron) &&
	    scr->posx >= p->winx &&
	    scr->posx < p->winx + brl.x && scr->posy >= p->winy &&
	    scr->posy < p->winy + brl.y)
	    brl.disp[(scr->posy - p->winy) * brl.x + scr->posx - p->winx] |=
		s_csrsize ? BIG_CSRCHAR : SMALL_CSRCHAR;
	brl.writebrl(&brl);
    }
}


/*
 * ----------------------------------------------------------------------
 * Load the low level Braille driver (a shared library).
 * ----------------------------------------------------------------------
 */
void init_braille(brl_drv_t *brl_drv)
{
    initbrl_func_t *initbrl;
    void           *lib_handle;
    char           *error;
    symbol_t       *sym;

    sym = lookup_symbol(brl_drv->scope, "driver");
    if (sym == NULL || sym->type != SY_STRINGVAR) {
	fprintf(stderr, "Braille driver library not specified\n");
	exit(5);
    }
    lib_handle = dlopen(sym->u.stringval, RTLD_LAZY);
    if (!lib_handle) {
	fprintf(stderr, "Can not load %s\n", sym->u.stringval);
        exit(5);
    }
    initbrl = dlsym(lib_handle, "initbrl");
    if ((error = dlerror()) != NULL)  {
	fprintf(stderr, "Can not resolve symbol: %s\n", error);
        exit(5);
    }

    if ((*initbrl)(&brl, brl_drv) != 0) {
	fprintf(stderr, "Braille driver initialization failed\n");
	close_screen();
	exit (5);
    }
    fwinshift = brl.x;
    hwinshift = fwinshift / 2;
    csr_offright = brl.x / 4;
    vwinshift = 5;
    fprintf(stderr, 
	    "Braille display has %d rows by %d cells and %d status cells\n",
	    brl.y, brl.x, brl.s);
    clrbrlstat();
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void init_speech(speech_drv_t *speech_drv)
{
    extern lang_open_f german_open, english_open;
    symbol_t *sym;
    lang_t   *english, *german;
    char     *str;
    int      i;

    /*
     * 0: LANG_BRITISH_ENGLISH
     * 1: LANG_GERMAN
     */
    no_speech_chans = 2;
    speech_chan = malloc(2 * sizeof(speech_chan_t));
    assert(speech_chan != NULL);

    sym = lookup_symbol(speech_drv->scope, "english");
    if (sym != NULL) {
	if (sym->type != SY_STRINGVAR) {
	    fprintf(stderr, "\"english\" must be string!\n");
	    exit(5);
	}
	str = sym->u.stringval;
    } else {
	printf("no synthesizer for english specified\n");
	str = NULL;
    }
    english = english_open(str);
    if (english == NULL) {
	fprintf(stderr, "Can not load synthesizer for english\n");
	exit(5);
    }
    speech_chan[0].lang  = english;
    speech_chan[0].langi = 0;

    sym = lookup_symbol(speech_drv->scope, "german");
    if (sym != NULL) {
	if (sym->type != SY_STRINGVAR) {
	    fprintf(stderr, "\"german\" must be string!\n");
	    exit(5);
	}
	str = sym->u.stringval;
    } else {
	printf("no synthesizer for german specified\n");
	str = NULL;
    }
    german = german_open(str);
    if (german == NULL) {
	fprintf(stderr, "Can not load synthesizer for german\n");
	exit(5);
    }
    speech_chan[1].lang  = german;
    speech_chan[1].langi = 1;
    /*
     * Set a default speech channel for all virtual terminals. This can
     * be overridden later, e.g. by the setup function.
     */
    for (i = 0; i < NBR_SCR; i++) {
	scrparam[i].speech = &speech_chan[0];
    }
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void close_speech(void)
{
    int i;
    
    for (i = 0; i < no_speech_chans; i++) {
	speech_chan[i].lang->close(speech_chan[i].lang);
    }
    free(speech_chan);
}


/*
 * ----------------------------------------------------------------------
 * Change the current terminal.
 * ----------------------------------------------------------------------
 */
void switchto(unsigned int scrno)
{
    curscr = scrno;

    if (scrno >= NBR_SCR)
	scrno = 1;
    p = &scrparam[scrno];
    usetable(p->dispmode ? TBL_ATTRIB : TBL_TEXT);
}


/*
 * ----------------------------------------------------------------------
 * Move the "visible" window of the Braille terminal around the screen.
 * Do checks for all border conditions.
 * ----------------------------------------------------------------------
 */
void setwinxy(scrstat_t *scr)
{
    if (s_slidewin) {
	/*
	 * Change position only if the coordonates are not already displayed
	 */
	if (scr->posx < p->winx || scr->posx >= p->winx + brl.x || 
	    scr->posy < p->winy || scr->posy >= p->winy + brl.y) {
	    p->winy = scr->posy < brl.y - 1 ? 0 : scr->posy - (brl.y - 1);
	    if (scr->posx < brl.x)
		p->winx = 0;
	    else if (scr->posx >= scr->cols - csr_offright)
		p->winx = scr->cols - brl.x;
	    else
		p->winx = scr->posx - (brl.x - csr_offright);
	}
    } else {
	if (scr->posx < p->winx || scr->posx >= p->winx + brl.x)
	    if (scr->posx >= (scr->cols / brl.x) * brl.x)
		p->winx = scr->cols - brl.x;
	    else
		p->winx = (scr->posx / brl.x) * brl.x;
	if (scr->posy < p->winy || scr->posy >= p->winy + brl.y)
	    if (scr->posy < brl.y - 1)
		p->winy = 0;
	    else
		p->winy = scr->posy - (brl.y - 1);
    }
}


/*
 * ----------------------------------------------------------------------
 * Write a plain message to the braille display.
 * Messages longer than the display are cut. I hope a display will have
 * at least 20 characters.
 * ----------------------------------------------------------------------
 */
void message(const char *s)
{
    int i, j, l;

    LogPrint(LOG_NOTICE, "message(%s)\n", s);
    usetable(TBL_TEXT);

    l = strlen (s);
    if (l <= brl.x * brl.y)
	j = l;
    else
	j = brl.x * brl.y - 1;
    /*
     * Do Braille translation using current table. Six-dot mode is
     * ignored, since case can be important, and blinking caps won't 
     * work. The part not used by the message is filled with blanks.
     */
    for (i = 0; i < j; i++)
	brl.disp[i] = curtbl[s[i]];
    for (; i < brl.x * brl.y; i++)
	brl.disp[i] = curtbl[' '];
    
    brl.writebrl(&brl);
    usetable(p->dispmode ? TBL_ATTRIB : TBL_TEXT);
}


/*
 * ----------------------------------------------------------------------
 * Clear status cells.
 * ----------------------------------------------------------------------
 */
void clrbrlstat(void)
{
    memset(brl.stat, 0, brl.s);
    brl.setbrlstat(&brl);
}


/*
 * ----------------------------------------------------------------------
 * Signal handler for unexpected signals.
 * ----------------------------------------------------------------------
 */
void error_handler(int signum)
{
    extern char *strsignal(int sig);

    LogPrint(LOG_ALERT, "brltty, received %s, terminating!\n",
	     strsignal(signum));
    exit(1);
}


/*
 * ----------------------------------------------------------------------
 * Catch SIGTERM, this gives a chance for cleanup.
 * ----------------------------------------------------------------------
 */
void termination_handler(int signum)
{
    keep_going = 0;
}


/*
 * ----------------------------------------------------------------------
 * Switch between text and attribute table.
 * ----------------------------------------------------------------------
 */
void usetable(int tbl)
{
    curtbl = (tbl == TBL_ATTRIB) ? attribtrans : texttrans;
}
