/*
 * BRLTTY - Access software for Unix for a blind person
 *          using a soft Braille terminal
 *
 * Copyright (C) 1995-1998 by The BRLTTY Team, All rights reserved.
 *
 *     Nicolas Pitre <nico@cam.org>
 *     Stphane Doyon <s.doyon@videotron.ca>
 *     Nikhil Nair <nn201@cus.cam.ac.uk>
 * This part by:
 *     Roger Butenuth <butenuth@uni-paderborn.de>
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 *
 * $Id: config_file.c,v 1.6 1999/03/11 21:08:29 butenuth Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include "config_file.h"

typedef struct {
    FILE *fp;
    char *name;
    int  line, column;
    int  ch;
} scanner_t;

struct {
    char   *name;
    tnum_t tnum;
} key_word[] = {
    { "braille", T_BRAILLE },
    { "braille_driver", T_BRAILLE_DRIVER },
    { "speech", T_SPEECH },
    { "speech_driver", T_SPEECH_DRIVER },
    { NULL, T_EOF }
};

const char *special_chars = "={}";

static int brl_drv_val_f(scanner_t *sc, token_t *t, symbol_t *sym);
static int speech_drv_val_f(scanner_t *sc, token_t *t, symbol_t *sym);


/*
 * ----------------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------------
 */
static int init_scanner(const char *name, scanner_t *sc)
{
    sc->name = strdup(name);
    assert(sc->name != NULL);

    sc->fp = fopen(name, "r");
    if (sc->fp == NULL) {
	free(sc->name);
	return errno;
    }
    sc->line   = 1;
    sc->column = 1;
    sc->ch = fgetc(sc->fp);

    return 0;
}


/*
 * ----------------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------------
 */
static int deinit_scanner(scanner_t *sc)
{
    fclose(sc->fp);
    free(sc->name);

    return 0;
}


/*
 * ----------------------------------------------------------------------------
 * Get the next character from the input stream.
 * ----------------------------------------------------------------------------
 */
static void get_ch(scanner_t *sc)
{
    if (sc->ch == '\n') {
	sc->line++;
	sc->column = 1;
    } else
	sc->column++;
    sc->ch = fgetc(sc->fp);
}


/*
 * ----------------------------------------------------------------------------
 * Skip input to the end of the current line (\n) or EOF.
 * ----------------------------------------------------------------------------
 */
static void skip_line(scanner_t *sc)
{
    while (sc->ch != EOF && sc->ch != '\n') {
	if (sc->ch == '\n') {
	    sc->line++;
	    sc->column = 1;
	} else
	    sc->column++;
	sc->ch = fgetc(sc->fp);
    }
}


/*
 * ----------------------------------------------------------------------------
 * Skip any whitespace in input, including comments.
 * ----------------------------------------------------------------------------
 */
static void skip_space(scanner_t *sc)
{
    while (sc->ch != EOF && (isspace(sc->ch) || sc->ch == '#')) {
	if (sc->ch == '#')
	    skip_line(sc);
	else {
	    if (sc->ch == '\n') {
		sc->line++;
		sc->column = 1;
	    } else
		sc->column++;
	    sc->ch = fgetc(sc->fp);
	}
    }
}

static int get_number(scanner_t *sc, token_t *t)
{
    fprintf(stderr, "get_number not implemented\n");
    exit(1);
}


/*
 * ----------------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------------
 */
static int get_string(scanner_t *sc, token_t *t)
{
    char *name;
    int  alen, len;

    len  = 0;
    alen = 100;
    name = malloc(alen);
    assert(name);
    get_ch(sc);			/* skip leading " */
    while (sc->ch != EOF && sc->ch != '"') {
	name[len++] = sc->ch;
	get_ch(sc);
	if (len == alen) {
	    alen += alen / 2;
	    name = realloc(name, alen);
	    assert(name);
	}
    }
    if (sc->ch != EOF)
	get_ch(sc);		/* skip trailing " */
    name[len] = '\0';
    name = realloc(name, strlen(name) + 1);
    if (name == NULL)
	return ENOMEM;

    t->tnum        = T_STRING;
    t->u.stringval = name;

    return 0;
}


/*
 * ----------------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------------
 */
static int get_symbol_or_keyword(scanner_t *sc, token_t *t)
{
    char *name;
    int  alen, len;
    int  i;

    len  = 0;
    alen = 100;
    name = malloc(alen);
    assert(name != NULL);

    do {
	name[len++] = sc->ch;
	get_ch(sc);
	if (len == alen) {
	    alen += alen / 2;
	    name = realloc(name, alen);
	    if (name == NULL)
		return ENOMEM;
	}
    } while (sc->ch != EOF && (isalnum(sc->ch) || sc->ch == '_'));
    name[len] = '\0';
    name = realloc(name, strlen(name) + 1);
    if (name == NULL)
	return ENOMEM;

    for (i = 0; key_word[i].name != NULL; i++)
	if (strcmp(key_word[i].name , name) == 0)
	    break;
    if (key_word[i].name != NULL) {
	t->tnum = key_word[i].tnum;
	free(name);
    } else {
	t->tnum     = T_SYMBOL;
	t->u.symval = name;
    }

    return 0;
}


/*
 * ----------------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------------
 */
static int get_sp_token(scanner_t *sc, token_t *t)
{
    int err = 0;

    switch (sc->ch) {
    case '=':
	t->tnum = T_EQUAL;
	break;
    case '{':
	t->tnum = T_LEFT_BRACE;
	break;
    case '}':
	t->tnum = T_RIGHT_BRACE;
	break;
    default:
	err = EBADMSG;
	break;
    }
    get_ch(sc);

    return err;
}

/*
 * ----------------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------------
 */
static int get_token(scanner_t *sc, token_t *t)
{
    int err = 0;

    skip_space(sc);
    t->line   = sc->line;
    t->column = sc->column;

    if (sc->ch == EOF) {
	t->tnum = T_EOF;
	return err;
    }
    if (isdigit(sc->ch))
	err = get_number(sc, t);
    else if (isalpha(sc->ch))
	err = get_symbol_or_keyword(sc, t);
    else if (sc->ch == '"')
	err = get_string(sc, t);
    else if (strchr(special_chars, sc->ch) != NULL)
	err = get_sp_token(sc, t);
    else {
	fprintf(stderr, "%s:%d: illegal character '%c'\n",
		sc->name, sc->line, sc->ch);
	get_ch(sc);
	err = EBADMSG;
    }

    return err;
}


/*
 * ----------------------------------------------------------------------------
 *  value ::= <string> | brl_drv | speech_drv
 * ----------------------------------------------------------------------------
 */
static int value_f(scanner_t *sc, token_t *t, symbol_t *sym)
{
    int err = 0;

    switch (t->tnum) {
    case T_STRING:
	sym->type = SY_STRINGVAR;
	sym->u.stringval = t->u.stringval;
	get_token(sc, t);
	break;
    case T_BRAILLE:
	err = brl_drv_val_f(sc, t, sym);
	break;
    case T_SPEECH:
	err = speech_drv_val_f(sc, t, sym);
	break;
    default:
	assert(0);
	break;
    }
    
    return err;
}


/*
 * ----------------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------------
 */
static symbol_t *make_symbol(token_t *t, scope_t *scope)
{
    symbol_t *node, **nodep;

    for (nodep = &scope->sym; *nodep != NULL; nodep = &(*nodep)->next)
	if (strcmp((*nodep)->name, t->u.symval) == 0)
	    break;
    if (*nodep) {
	node = *nodep;
    } else {
	node = malloc(sizeof(symbol_t));
	assert(node != NULL);
	
	node->next  = NULL;
	*nodep = node;
	
	node->line  = t->line;
	node->name  = t->u.symval;
	node->scope = scope;
    }

    return node;
}


/*
 * ----------------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------------
 */
static scope_t *make_scope(scope_t *outer_scope)
{
    scope_t *new_scope;

    new_scope = malloc(sizeof(scope_t));
    assert(new_scope != NULL);

    new_scope->pred = outer_scope;
    new_scope->sym  = NULL;

    return new_scope;
}


/*
 * ----------------------------------------------------------------------------
 * assignment ::= <symbol> "=" value
 * ----------------------------------------------------------------------------
 */
static int assignment_f(scanner_t *sc, token_t *t, scope_t *scope)
{
    symbol_t *sym;
    int      err = 0;

    assert(t->tnum == T_SYMBOL);
    sym = make_symbol(t, scope);
    get_token(sc, t);	/* skip symbol */
    if (t->tnum != T_EQUAL) {
	fprintf(stderr, "%s:%d: '=' expected\n", sc->name, t->line);
	return EBADMSG;
    }
    get_token(sc, t);	/* skip "=" */
    err = value_f(sc, t, sym);
    return err;
}


/*
 * ----------------------------------------------------------------------------
 * assign_list ::= [assignment]
 * ----------------------------------------------------------------------------
 */
static int assign_list_f(scanner_t *sc, token_t *t, scope_t *scope)
{
    int      err = 0;

    while (!err && t->tnum == T_SYMBOL) {
	err = assignment_f(sc, t, scope);
    }

    return err;
}


/*
 * ----------------------------------------------------------------------------
 * brl_drv_val ::= "braille" "{" assign_list "}"
 * ----------------------------------------------------------------------------
 */
static int brl_drv_val_f(scanner_t *sc, token_t *t, symbol_t *sym)
{
    int       err = 0;
    brl_drv_t *brl_drv;

    assert(t->tnum == T_BRAILLE);
    get_token(sc, t);		/* skip "braille" */
    if (t->tnum != T_LEFT_BRACE) {
	fprintf(stderr, "%s:%d: '{' expected\n", sc->name, t->line);
	return EBADMSG;
    }
    get_token(sc, t);		/* skip "{" */

    brl_drv = malloc(sizeof(brl_drv_t));
    assert(brl_drv != NULL);

    brl_drv->scope = make_scope(sym->scope);
    if (t->tnum != T_SYMBOL) {
	fprintf(stderr, "%s:%d: symbol expected\n", sc->name, t->line);
	return EBADMSG;
    }
    err = assign_list_f(sc, t, brl_drv->scope);
    if (!err) {
	sym->type      = SY_BRL_DRV;
	sym->u.brl_drv = brl_drv;

	if (t->tnum != T_RIGHT_BRACE) {
	    fprintf(stderr, "%s:%d: '}' expected\n", sc->name, t->line);
	    return EBADMSG;
	}
	get_token(sc, t);	/* skip "}" */
    }

    return err;
}


/*
 * ----------------------------------------------------------------------------
 * speech_drv_val ::= "speech" "{" assign_list "}"
 * ----------------------------------------------------------------------------
 */
static int speech_drv_val_f(scanner_t *sc, token_t *t, symbol_t *sym)
{
    int       err = 0;
    speech_drv_t *speech_drv;

    assert(t->tnum == T_SPEECH);
    get_token(sc, t);		/* skip "speech" */
    if (t->tnum != T_LEFT_BRACE) {
	fprintf(stderr, "%s:%d: '{' expected\n", sc->name, t->line);
	return EBADMSG;
    }
    get_token(sc, t);		/* skip "{" */

    speech_drv = malloc(sizeof(speech_drv_t));
    assert(speech_drv != NULL);

    speech_drv->scope = make_scope(sym->scope);
    if (t->tnum != T_SYMBOL) {
	fprintf(stderr, "%s:%d: symbol expected\n", sc->name, t->line);
	return EBADMSG;
    }
    err = assign_list_f(sc, t, speech_drv->scope);
    if (!err) {
	sym->type      = SY_SPEECH_DRV;
	sym->u.speech_drv = speech_drv;

	if (t->tnum != T_RIGHT_BRACE) {
	    fprintf(stderr, "%s:%d: '}' expected\n", sc->name, t->line);
	    return EBADMSG;
	}
	get_token(sc, t);	/* skip "}" */
    }

    return err;
}


/*
 * ----------------------------------------------------------------------------
 * braille_driver ::= "braille_driver" "=" <symbol>
 * ----------------------------------------------------------------------------
 */
static int braille_driver_f(scanner_t *sc, token_t *t, syntax_tree_t *tree)
{
    int err = 0;
    symbol_t *sym;

    assert(t->tnum == T_BRAILLE_DRIVER);
    get_token(sc, t);		/* skip "braille-driver" */
    if (t->tnum != T_EQUAL) {
	fprintf(stderr, "%s:%d: '=' expected\n", sc->name, t->line);
	return EBADMSG;
    }
    get_token(sc, t);		/* skip "=" */
    if (t->tnum != T_SYMBOL) {
	fprintf(stderr, "%s:%d: symbol expected\n", sc->name, t->line);
	return EBADMSG;
    }
    sym = lookup_symbol(tree->scope, t->u.symval);
    if (sym == NULL) {
	fprintf(stderr, "%s:%d: symbol %s undefined\n",
		sc->name, t->line, t->u.symval);
	return EBADMSG;
    }
    if (sym->type != SY_BRL_DRV) {
	fprintf(stderr, "%s:%d: %s is not a Braille driver\n",
		sc->name, t->line, t->u.symval);
	return EBADMSG;
    }
    tree->brl_drv = sym->u.brl_drv;
    get_token(sc, t);		/* skip symbol */

    return err;
}


/*
 * ----------------------------------------------------------------------------
 * speech_driver ::= "speech_driver" "=" <symbol>
 * ----------------------------------------------------------------------------
 */
static int speech_driver_f(scanner_t *sc, token_t *t, syntax_tree_t *tree)
{
    int err = 0;
    symbol_t *sym;

    assert(t->tnum == T_SPEECH_DRIVER);
    get_token(sc, t);		/* skip "speech-driver" */
    if (t->tnum != T_EQUAL) {
	fprintf(stderr, "%s:%d: '=' expected\n", sc->name, t->line);
	return EBADMSG;
    }
    get_token(sc, t);		/* skip "=" */
    if (t->tnum != T_SYMBOL) {
	fprintf(stderr, "%s:%d: symbol expected\n", sc->name, t->line);
	return EBADMSG;
    }
    sym = lookup_symbol(tree->scope, t->u.symval);
    if (sym == NULL) {
	fprintf(stderr, "%s:%d: symbol %s undefined\n",
		sc->name, t->line, t->u.symval);
	return EBADMSG;
    }
    if (sym->type != SY_SPEECH_DRV) {
	fprintf(stderr, "%s:%d: %s is not a Speech driver\n",
		sc->name, t->line, t->u.symval);
	return EBADMSG;
    }
    tree->speech_drv = sym->u.speech_drv;
    get_token(sc, t);		/* skip symbol */

    return err;
}


/*
 * ----------------------------------------------------------------------------
 * item ::= assignment | braille_driver | speech_driver
 * ----------------------------------------------------------------------------
 */
static int item_f(scanner_t *sc, token_t *t, syntax_tree_t *tree)
{
    int    err = 0;

    switch (t->tnum) {
    case T_BRAILLE_DRIVER:
	err = braille_driver_f(sc, t, tree);
	break;
    case T_SPEECH_DRIVER:
	err = speech_driver_f(sc, t, tree);
	break;
    case T_SYMBOL:
	err = assignment_f(sc, t, tree->scope);
	break;
    default:
	fprintf(stderr, "%s:%d: top level, bad item\n", sc->name, t->line);
	err = EBADMSG;
	break;
    }

    return err;
}


/*
 * ----------------------------------------------------------------------------
 *
 * Returns zero or a positive error number.
 * ----------------------------------------------------------------------------
 */
int parse_config_file(const char *file, syntax_tree_t **tree_p)
{
    scanner_t     sc;
    token_t       t;
    syntax_tree_t *tree = NULL;
    int           err, err2;

    err = err2 = init_scanner(file, &sc);
    if (!err) {
	tree = malloc(sizeof(syntax_tree_t));
	assert(tree != NULL);
	tree->filename = strdup(file);
	assert(tree->filename != NULL);
	tree->scope = malloc(sizeof(scope_t));
	assert(tree->scope != NULL);
	tree->scope->pred = NULL;
	tree->scope->sym  = NULL;
	tree->brl_drv    = NULL;
	tree->speech_drv = NULL;
    }

    if (!err)
	get_token(&sc, &t);
    while (!err && t.tnum != T_EOF) {
	err = item_f(&sc, &t, tree);
    }
    
    if (!err2)
	err2 = deinit_scanner(&sc);
    if (!err && err2)
	err = err2;
    if (!err)
	*tree_p = tree;

    if (err)
	errno = err;
    return err;
}


/*
 * ----------------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------------
 */
symbol_t *lookup_symbol(scope_t *scope, const char *name)
{
    symbol_t *sym = NULL;

    while (scope != NULL) {
	for (sym = scope->sym; sym != NULL; sym = sym->next) {
	    if (strcmp(name, sym->name) == 0)
		return sym;
	}
	scope = scope->pred;
    }
    return NULL;
}

#if 0
/* for testing only */

void dump_brl_drv_val(int level, brl_drv_t *brl);
void dump_speech_drv_val(int level, speech_drv_t *speech);

void indent(int level)
{
    int i;
    
    for (i = 0; i < level; i++)
	printf("    ");
}

void dump_sym(int level, symbol_t *sym)
{
    indent(level); printf("%s = ", sym->name);
    switch (sym->type) {
    case SY_STRINGVAR:
	printf("\"%s\"\n", sym->u.stringval);
	break;
    case SY_BRL_DRV:
	dump_brl_drv_val(level, sym->u.brl_drv);
	break;
    case SY_SPEECH_DRV:
	dump_speech_drv_val(level, sym->u.speech_drv);
	break;
    default:
	printf("unknown symbol type\n");
	break;
    }
}

void dump_scope(int level, scope_t *scope)
{
    symbol_t *sym;

    printf("{ # %p <- %p\n", scope->pred, scope);
    for (sym = scope->sym; sym; sym = sym->next) {
	dump_sym(level + 1, sym);
    }
    indent(level); printf("}\n");
}

void dump_brl_drv_val(int level, brl_drv_t *brl)
{
    printf("braille ");
    dump_scope(level, brl->scope);
}

void dump_speech_drv_val(int level, speech_drv_t *speech)
{
    printf("speech ");
    dump_scope(level, speech->scope);
}

void dump_item(int level, item_t *item)
{
    switch (item->type) {
    default:
	printf("unknown item\n");
	break;
    }
}

void dump_syntax_tree(syntax_tree_t *tree)
{
/*    item_t *item;*/

    printf("global_scope = ");
    dump_scope(0, tree->scope);
    printf("Activated Braille driver = ");
    dump_scope(0, tree->brl_drv->scope);
    printf("Activated Speech driver = ");
    dump_scope(0, tree->speech_drv->scope);
#if 0
    printf("item list:\n");
    for (item = tree->first; item; item = item->next)
	dump_item(0, item);
#endif
}


int main(int argc, char *argv[])
{
    syntax_tree_t *tree;
    int res;

    res = parse_config_file("test-config", &tree);
    if (res) {
	perror("parse");
    }

    if (!res)
	dump_syntax_tree(tree);

    return 0;
}
#endif
