/*
 * BRLTTY - Access software for Unix for a blind person
 *          using a soft Braille terminal
 *
 * Copyright (C) 1995-1998 by The BRLTTY Team, All rights reserved.
 *
 *     Nicolas Pitre <nico@cam.org>
 *     Stphane Doyon <s.doyon@videotron.ca>
 *     Nikhil Nair <nn201@cus.cam.ac.uk>
 * This part completely rewritten by:
 *     Roger Butenuth <butenuth@uni-paderborn.de>
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/termios.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/vt.h>
#include <errno.h>
#include <sched.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#if KBD_SNIFF
#include <linux/kbd_sniff.h>
#endif

#include "brl.h"
#include "misc.h"
#include "scr.h"
#include "config.h"


#define ROUTING_TIMEOUT 2000	/* in milliseconds */

static void live_getstat(scrstat_t *stat);
static unsigned char *live_getscr(winpos_t pos, unsigned char *buffer,
				  short mode);

static int  froz_open(void);
static void froz_close(void);
static void froz_getstat(scrstat_t *stat);
static unsigned char *froz_getscr(winpos_t pos, unsigned char *buffer,
				  short mode);


typedef struct {
    int           initialized;
    scrstat_t     stat;
    unsigned char *text;
    unsigned char *attrib;
    void          (*getstat)(scrstat_t *stat);
    unsigned char *(*getscr)(winpos_t pos, unsigned char *buffer, short mode);
} screen_t;

screen_t screens[2];		/* 0 = live, 1 = frozen */

screen_t *current = NULL;

static int vcsa_fd, cons_fd;
#if KBD_SNIFF
static int kbd_sniffing = 0;
#endif

/*
 * Variables and definitions for cut'n'paste.
 */

static int cut_begx = 0, cut_begy = 0;

/*
 * The following code is not my fault, but somebody wanted to save three
 * byte and forgot correct padding in the structure. Arrrrgh!
 */
#define XS (*(short*)(&sel.subcode + 1 + 0 * sizeof(short)))
#define YS (*(short*)(&sel.subcode + 1 + 1 * sizeof(short)))
#define XE (*(short*)(&sel.subcode + 1 + 2 * sizeof(short)))
#define YE (*(short*)(&sel.subcode + 1 + 3 * sizeof(short)))
#define MODE (*(short*)(&sel.subcode + 1 + 4 * sizeof(short)))

struct {
    char  subcode;
    short dummy[5];
} sel;


/*
 * ----------------------------------------------------------------------
 * Initialize screen reading, return 0 on success.
 *
 * Initialize keyboard sniffing. Magic key combination is
 * "left shift + right shift" to enter filter mode, "space" to leave
 * filter mode. In filter mode all keypresses are not given to other
 * application, only the sniffer gets this keypresses unless filter
 * mode is left.
 * ----------------------------------------------------------------------
 */
int init_screen(void)
{
#if KBD_SNIFF
    struct sniff_ioctl_set set;
#endif
    int                    i;

    for (i = 0; i < 2; i++) {
	screens[i].initialized = 0;
	screens[i].text        = NULL;
	screens[i].attrib      = NULL;
    }
    screens[LIVE_SCRN].getstat = live_getstat;
    screens[LIVE_SCRN].getscr  = live_getscr;
    screens[FROZ_SCRN].getstat = froz_getstat;
    screens[FROZ_SCRN].getscr  = froz_getscr;

    vcsa_fd = open(VCSADEV, O_RDONLY);
    if(vcsa_fd<0) {
	LogAndStderr(LOG_WARNING,"Can't open vcsa device '%s': %s\n",
		     VCSADEV, strerror(errno));
	return 1;
    }
    cons_fd = open (CONSOLE, O_RDONLY);
    if (cons_fd < 0) {
	LogAndStderr(LOG_WARNING, "Can't open console device '%s': %s\n",
		     CONSOLE, strerror(errno));
	close (vcsa_fd);
	return 1;
    }
    screens[0].initialized = 1;
    current = &screens[LIVE_SCRN];

#if KBD_SNIFF
    set.hot_key_enable = 1;
    set.nr_hot_keys = 2;
    set.hot_key[0]  = 42;               /* left shift */
    set.hot_key[1]  = 54;               /* right shift */
    set.disable_key = 57;               /* space */

    if (ioctl(vcsa_fd, KBD_SNIFF_SET, &set) < 0) {
	fprintf(stderr, "no keyboard sniffing in kernel\n");
	kbd_sniffing = 0;
    } else {
	fprintf(stderr, "keyboard sniffing on\n");
	kbd_sniffing = 1;
    }
#endif

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * Disable sniffing mode, close files.
 * ----------------------------------------------------------------------
 */
void close_screen(void)
{
#if KBD_SNIFF
    struct sniff_ioctl_set set;

    if (kbd_sniffing) {
	set.hot_key_enable = 0;
	if (ioctl(vcsa_fd, KBD_SNIFF_SET, &set) < 0)
	    perror("ioctl set");
    }
#endif

    close(cons_fd);
    close(vcsa_fd);
}


/*
 * ----------------------------------------------------------------------
 * Get screen size and cursor position of current screen and
 * current active console
 * ----------------------------------------------------------------------
 */
void getstat(scrstat_t *stat)
{
    assert(current->initialized);

    current->getstat(stat);
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
unsigned char *getscr(winpos_t pos, unsigned char *buffer, short mode)
{
    assert(current->initialized);

    return current->getscr(pos, buffer, mode);
}


/*
 * ----------------------------------------------------------------------
 * disp = LIVE_SCRN or FROZ_SCRN.
 * ----------------------------------------------------------------------
 */
int selectdisp(int disp)
{
    static int old_disp = LIVE_SCRN;

    if (!(old_disp & FROZ_SCRN) && (disp & FROZ_SCRN)) {
	/*
	 * Freeze (copy live screen to backing store). When freeze fails
	 * (e.g. not enough memory), ignore the request.
	 */
	if (froz_open())
	    disp &= ~FROZ_SCRN;
    }
    if ((old_disp & FROZ_SCRN) && !(disp & FROZ_SCRN)) {
	froz_close();		/* never fails */
    }
    /*
     * Select current screen.
     */
    current = &screens[LIVE_SCRN];
    if (disp & FROZ_SCRN)
	current = &screens[FROZ_SCRN];

    old_disp = disp;
    return disp;
}


/*
 * Conversion from ibmpc==cp437 to iso-latin-1 for chars>=128
 * Note that this is a quick hack until we can get all this directly
 * from the kernel console.  We also need to find how to know if 
 * such a translation table is in use.
 */
static unsigned char charset_conv[128] = {
    199, 252, 233, 226, 228, 224, 229, 231, 234, 235, 232, 239, 238, 236,
    196, 197, 201, 230, 198, 244, 246, 242, 251, 249, 255, 214, 220, 162,
    163, 165, 158, 159, 225, 237, 243, 250, 241, 209, 170, 186, 191, 169,
    172, 189, 188, 161, 171, 187, 248, 164, 253, 179, 180, 145,  20, 156,
    184, 185,  21, 175, 166, 174, 190, 168, 192, 193, 194, 195, 142, 143,
    146, 128, 200, 144, 202, 203, 204, 205, 206, 207, 208, 157, 210, 211,
    212, 213, 153, 215, 216, 217, 218, 219, 154, 221, 222, 152, 133, 223,
    131, 227, 132, 134, 181, 135, 138, 130, 136, 137, 141, 173, 140, 139,
    240, 177, 149, 155, 147, 245, 247, 148, 176, 151, 183, 150, 129, 178,
    254, 160
};


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
static void live_getstat(scrstat_t *stat)
{
    unsigned char buffer[4];
    struct vt_stat vtstat;
    
    lseek(vcsa_fd, 0, SEEK_SET);	/* go to start of file */
    read(vcsa_fd, buffer, 4);		/* get screen status bytes */
    stat->rows = buffer[0];
    stat->cols = buffer[1];
    stat->posx = buffer[2];
    stat->posy = buffer[3];
    ioctl(cons_fd, VT_GETSTATE, &vtstat);
    stat->no = vtstat.v_active;
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
unsigned char *live_getscr(winpos_t pos, unsigned char *buffer, short mode)
{
    scrstat_t     stat;		/* screen statistics */
    unsigned char *linebuf;
    off_t         start;
    size_t        linelen;
    int           i, j;
    unsigned char *dst;

    getstat(&stat);
    if (pos.left < 0 || pos.top < 0 || pos.width < 1 || pos.height < 1
	||mode < 0 || mode > 1 || pos.left + pos.width > stat.cols
	||pos.top + pos.height > stat.rows)
	return NULL;
    /* start offset */
    start = 4 + (pos.top * stat.cols + pos.left) * 2;
    /* number of bytes to read for one complete line */
    linelen = 2 * pos.width;

    linebuf = malloc(linelen);
    if (linebuf == NULL)
	return NULL;

    assert(buffer != NULL);
    dst = buffer;
    for (i = 0; i < pos.height; i++) {
	lseek (vcsa_fd, start + i * stat.cols * 2, SEEK_SET);
	read (vcsa_fd, linebuf, linelen);
	for (j = 0; j < pos.width; j++) {
	    if (mode == SCR_TEXT) {
		/*
		 * Conversion from ibmpc==cp437 to iso-latin-1 for chars >= 128
		 */
		if (linebuf[2 * j] >= 128)
		    *dst = charset_conv[linebuf[2 * j] - 128];
		else
		    *dst = linebuf[2 * j];
	    } else /* mode == SCR_ATTRIB */ { 
		*dst = linebuf[2 * j + 1];
	    }
	    dst++;
	}
    }
    free(linebuf);

    return buffer;
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
int froz_open(void)
/* (Screen *src) */
{
#if 0
    src->getstat(stat);
    if (!(text = new unsigned char[stat.rows * stat.cols]))
	return 1;
    if (!(attrib = new unsigned char[stat.rows * stat.cols])) {
	delete text;
	text = 0;
	return 1;
    }
    if (!src->getscr ((winpos) { 0, 0, stat.cols, stat.rows }, text, SCR_TEXT)
	|| !src->getscr ((winpos) { 0, 0, stat.cols, stat.rows }, attrib,
	SCR_ATTRIB)) {
	delete text;
	text = 0;
	delete attrib;
	return 2;
    }
#endif
    return 0;
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void froz_close(void)
{
#if 0
    if (text) {
	free(text);
	free(attrib);
    }
#endif
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void froz_getstat (scrstat_t *stat2)
{
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
unsigned char *froz_getscr(winpos_t pos, unsigned char *buffer, short mode)
{
#if 0
    unsigned char *scrn;

    if (pos.left < 0 || pos.top < 0 || pos.width < 1 || pos.height < 1
      || mode < 0 || mode > 1 || pos.left + pos.width > stat.cols
	|| pos.top + pos.height > stat.rows)
	return NULL;
    scrn = (mode == SCR_TEXT) ? text : attrib;
    for (int i = 0; i < pos.height; i++)
	memcpy (buffer + i * pos.width, scrn + (pos.top + i)* stat.cols +
		pos.left, pos.width);
#endif
    return buffer;
}


/*
 * ----------------------------------------------------------------------
 * Set model name for the right help file. 
 * I do not really like this function, but I do not want to change
 * the interface to the low level braille driver.
 * ----------------------------------------------------------------------
 */
void sethlpscr(int model)
{
    ;				/* currently ignored */
}


#if KBD_SNIFF
/*
 *----------------------------------------------------------------------
 * Convert given shift state and keycode to ascii. 
 * Return 0 whenever there is no meaningful translation.
 *----------------------------------------------------------------------
 */
int key_to_ascii(int shift_state, int keycode)
{
    struct kbentry key;
    int            type, sym;
    int            ascii = 0;

    if (keycode & 0x80)
        return 0;               /* ignore key release */

    key.kb_table = shift_state & 0xff;
    key.kb_index = keycode;
    key.kb_value = 0;

    if (ioctl(cons_fd, KDGKBENT, &key) < 0) {
        perror("get key entry ioctl");
        return 0;
    }
    if (key.kb_value != K_HOLE) {
        type = KTYP(key.kb_value) & 0x0f;
        sym  = KVAL(key.kb_value);
        switch (type) {
        case KT_LETTER:
            if ((shift_state & SNIFF_CAPSLOCK) != 0)
                ascii = sym - 'a' + 'A';
            else
                ascii = sym;
            break;
        case KT_LATIN:
            ascii = sym;
            break;
        case KT_ASCII:
        case KT_FN:
        case KT_SPEC:
        case KT_PAD:
        case KT_DEAD:
        case KT_CONS:
        case KT_CUR:
        case KT_SHIFT:
        case KT_META:
        case KT_LOCK:
        case KT_SLOCK:
            break;
        }
    }
    return ascii;
}


/*
 *----------------------------------------------------------------------
 * Get one keypress. This can be done blocking or with polling.
 * keycode == 0 means no key available (polling only)
 *----------------------------------------------------------------------
 */
int get_key(int blocking, int *shift_state, int *keycode)
{
    static struct sniff_ioctl_get get;
    static int                    pos = 0;
    int                           result = 0;

    if (pos == get.count) {
	/*
	 * buffer is empty
	 */
	pos = 0;
	get.blocking = blocking;
	get.count = 0;
	result = ioctl(vcsa_fd, KBD_SNIFF_GET, &get);
	if (result < 0)
	    return result;
    }

    if (pos < get.count) {
	*shift_state = get.event[pos].shift_state;
	*keycode     = get.event[pos].keycode;
	pos++;
    } else {
	*shift_state = 0;
	*keycode     = 0;
    }

    return result;
}


/*
 * ----------------------------------------------------------------------
 * Take a keypress/keyrelease as input and convert it into an braille
 * command.
 * ----------------------------------------------------------------------
 */
int build_command(int shift_state, int keycode)
{
    int           cmd = EOF;
    unsigned char ascii;

    if  ((keycode & 0x80) != 0) {
 	/*
	 * key release, ignore
	 */
    } else {
	/*
	 * key press
	 */
	ascii = key_to_ascii(shift_state, keycode);
	switch (keycode) {
	case 103:		/* cursor up */
	    cmd = CMD_LNUP;
	    break;
	case 108:		/* cursor down */
	    cmd = CMD_LNDN;
	    break;
	case 106:		/* cursor right */
	    cmd = CMD_HWINRT;
	    break;
	case 105:		/* cursor left */
	    cmd = CMD_HWINLT;
	    break;
	case 1:			/* Escape */
	    cmd = CMD_CONFMENU;
	    break;
	default:
	}
    }
    return cmd;
}
#endif


/*
 * ----------------------------------------------------------------------
 * Get an input event. When keyboard sniffing is active, try keyboard
 * first. Whenever there is no event, try braille terminal.
 * ----------------------------------------------------------------------
 */
int get_input_event(int type, int blocking)
{
    int result = EOF;
    
    do {
#if KBD_SNIFF
	if (kbd_sniffing) {
	    int shift_state, keycode;

	    (void)get_key(0, &shift_state, &keycode);
	    
	    if (keycode != 0) {
		if (shift_state & SNIFF_HOT_MODE) {
		    result = build_command(shift_state, keycode);
		}
	    }
	}
#endif
	if (result == EOF)
	    result = brl.readbrl(&brl, type);
	/*
	 * In blocking mode, wait a little, otherwise we would burn
	 * up cpu power. This solution is not very elegant.
	 */
	if (result == EOF && blocking == EVNT_BLOCKING)
	    usleep(20000);
    } while (result == EOF && blocking == EVNT_BLOCKING);

    return result;
}


/*
 * ----------------------------------------------------------------------
 * Wait, until the cursor moves from the specified position.
 * This is busy waiting, so be careful: If the timeout specified is
 * zero, do at least an sched_yield(). This should enable the active
 * application to do some work.
 * We wait no longer than one second in total.
 * Return 0 on success, 1 on timeout or change of screen.
 * ----------------------------------------------------------------------
 */
static int wait_for_move(int screen, int x, int y)
{
    scrstat_t scr;
    int       count;

    LogPrint(LOG_NOTICE, "for move from %d %d\n", x, y);
    count = 1000;
    do {
#if 1
	sched_yield();
#else
	delay(2);
#endif
	live_getstat(&scr);
	count--;
    } while (count > 0 && scr.no == screen &&
	     scr.posx == x && scr.posy == y);

    LogPrint(LOG_NOTICE, "pos is %d %d count is %d\n", scr.posx, scr.posy, count);
    return (count == 0 || scr.no != screen);
}



/*
 * ----------------------------------------------------------------------
 * Try to move the cursor on the screen to the specified position.
 * A -1 as x means only to adjust the line, not the column.
 * ----------------------------------------------------------------------
 */
void cursor_routing(int curscr, int x, int y)
{
    struct timeval start, current;
    int            wait_fail;
    int            dif;
    scrstat_t      scr;
    int            cur_cons_fd = -1;

    LogPrint(LOG_NOTICE, ">>> routing %d %d\n", x, y);
    cur_cons_fd = open (CONSOLE, O_RDONLY);
    assert(cur_cons_fd >= 0);
    gettimeofday(&start, NULL);
    gettimeofday(&current, NULL);
    /*
     * Deal with vertical movement first, ignoring horizontal jumping ...
     */
    live_getstat(&scr);
    dif = y - scr.posy;
    wait_fail = 0;
    LogPrint(LOG_NOTICE, " curscr = %d scr.no = %d\n", curscr, scr.no);
    while (dif != 0 && curscr == scr.no && wait_fail < 3 &&
	   elapsed_msec(&start, &current) < ROUTING_TIMEOUT) {
	live_getstat(&scr);
	inskey(cur_cons_fd, dif > 0 ? DN_CSR : UP_CSR);
	if (wait_for_move(curscr, scr.posx, scr.posy))
	    wait_fail++;
	
	live_getstat(&scr);
	dif = y - scr.posy;
	gettimeofday(&current, NULL);
    }
    
    if(x != -1) {
	/*
	 * don't do this for vertical-only routing (x=-1)
	 * Now horizontal movement, quitting if the vertical position
	 * is wrong:
	 */
	live_getstat(&scr);
	dif = x - scr.posx;
	wait_fail = 0;
	while (dif != 0 && scr.posy == y && wait_fail < 3 &&
	       curscr == scr.no &&
	       elapsed_msec(&start, &current) < ROUTING_TIMEOUT) {
	    live_getstat(&scr);
	    inskey(cur_cons_fd, dif > 0 ? RT_CSR : LT_CSR);
	    if (wait_for_move(curscr, scr.posx, scr.posy))
		wait_fail++;
	    LogPrint(LOG_NOTICE, "%s\n", dif > 0 ? "right" : "left");

	    live_getstat(&scr);
	    dif = x - scr.posx;
	    gettimeofday(&current, NULL);
	}
    }

    if (elapsed_msec(&start, &current) >= ROUTING_TIMEOUT)
	LogPrint(LOG_NOTICE, "    timout fail, %d ms\n",
		 elapsed_msec(&start, &current));
    close(cur_cons_fd);
    LogPrint(LOG_NOTICE, "<<< routing wait_fail = %d\n", wait_fail);
}


/*
 * ----------------------------------------------------------------------
 * The ultimate goal for this function is to be able to insert characters,
 * as well as arrow key controls, even if the keyboard on the console is
 * in K_RAW mode (as in dosemu particularly).  
 * For instance, at least, the insertion is not performed if in 
 * raw keyboard to prevent garbage results.
 * ----------------------------------------------------------------------
 */
void inskey(int fd, unsigned char *string)
{
    long kbmode;

    ioctl(fd, KDGKBMODE, &kbmode);
    if(kbmode == K_XLATE) {
	while (*string) {
	    if (ioctl(fd, TIOCSTI, string))
		break;
	    string++;
	}
    }
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void cut_begin(int x, int y)
{
    cut_begx = x;
    cut_begy = y;
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void cut_end(int x, int y)
{
    int         fd;
    
    sel.subcode = 2;		/* cut */
    MODE = 0;			/* character by character */
    XS = cut_begx;
    YS = cut_begy;
    XE = x;
    YE = y;
    fd = open(CONSOLE, O_RDONLY);
    assert(fd >= 0);
    ioctl(fd, TIOCLINUX, &sel);
    close(fd);
}


/*
 * ----------------------------------------------------------------------
 *
 * ----------------------------------------------------------------------
 */
void cut_paste(void)
{
    int         fd;
    
    sel.subcode = 3;		/* paste */
    fd = open(CONSOLE, O_RDONLY);
    assert(fd >= 0);
    ioctl(fd, TIOCLINUX, &sel);
    close(fd);
}
