/*
 * BRLTTY - Access software for Unix for a blind person
 *          using a soft Braille terminal
 *
 * Copyright (C) 1998 by The BRLTTY Team, All rights reserved.
 *
 * This part from:
 *     Roger Butenuth <butenuth@uni-paderborn.de>
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 *
 * $Id: setup.c,v 1.12 1999/03/11 21:08:30 butenuth Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "config.h"
#include "brl.h"
#include "scr.h"
#include "misc.h"
#include "Speech/language.h"
#include "Speech/synthesizer.h"
#include "setup.h"

#define SETUP_LANGUAGES 2	/* english, german */

static menu_t *root;

static menu_t *menu;
static int    keep_going;

typedef struct language_struct {
    int            langi;	/* language index */
    brltty_param_t *p;		/* screen this language entry controls */
    entry_t        *entry;	/* pointer entry this belongs to */
} language_t;

typedef struct speech_par_struct {
    int            is_lang;	/* language parameter? (synth otherwise) */
    union {
	synth_t *synth;
	lang_t  *lang;
    } u;
    int            par;		/* parameter to change */
    entry_t        *entry;	/* pointer entry this belongs to */
} speech_par_t;

static void   dump_tree(FILE *fp, menu_t *menu, int level, int lang);
static void   exit_fun(void *this, cmd_e cmd, msg_t *text, int *value);
static void   binary_fun(void *this, cmd_e cmd, msg_t *text, int *value);
static void   integer_fun(void *this, cmd_e cmd, msg_t *text, int *value);
static node_t *make_braille_menu(node_t *parent);
static node_t *make_speech_menu(node_t *parent);

static msg_t yes_text  = { { "yes", "ja"   } };
static msg_t no_text   = { { "no",  "nein" } };
static msg_t on_text   = { { "on",  "ein"  } };
static msg_t off_text  = { { "off", "aus"  } };

static msg_t lang_names[SETUP_LANGUAGES] = {
    { { "British English", "Englisch" } },
    { { "German", "Deutsch" } }
};

language_t langs[NBR_SCR];


/*
 * ----------------------------------------------------------------------
 * Build the setup tree.
 * ----------------------------------------------------------------------
 */
void init_setup(void)
{
    static node_t exit_entry;
    static msg_t  root_title = { { "Setup", "Einstellungen" } };
    static msg_t  exit_name = { { "Exit", "Verlassen" } };
    exit_entry.type = N_ENTRY;
    
    exit_entry.e.name    = &exit_name;
    exit_entry.e.fun     = exit_fun;
    exit_entry.e.private = NULL;

    root = make_menu(&root_title, &exit_entry);

    add_to_menu(root, make_braille_menu((node_t*)root));
    add_to_menu(root, make_speech_menu((node_t*)root));

    menu = root;
    keep_going = 1;
    menu->cur = 0;
}


/*
 * ----------------------------------------------------------------------
 * Perform setup until user leaves the setup tree.
 * ----------------------------------------------------------------------
 */
void setup(int curscr, brltty_param_t *p)
{
    lang_t  *lang;
    synth_t *synth;
    int     langi;
    char    text[80];
    int     len, mlen;
    node_t  *node;
    int     val;
    msg_t   title;
    int     event;

    lang  = p[curscr].speech->lang;
    synth = p[curscr].speech->lang->synth;
    langi = p[curscr].speech->langi;

    keep_going = 1;
    while (keep_going) {
	/*
	 * Build the text of the menu. On the braille display this is
	 * the name of the menu followed by the name of the submenu
	 * or the string generated by the entry, e.g "volume = 50".
	 * In speech the name of the menu is omitted for brevity.
	 */
	sprintf(text, "%s ", menu->title->txt[langi]);
	mlen = strlen(text);
	len = mlen;
	node  = menu->entry[menu->cur];
	switch (node->type) {
	case N_MENU:
	    title = *node->m.title;
	    break;
	case N_ENTRY:
	    node->e.fun(node->e.private, C_GET_VALUE, &title, &val);
	    break;
	}
	len += sprintf(text + len, "%s", title.txt[langi]);
	/*
	 * Display the full text on the display, but speak only the second
	 * part, except we are on top of the menu.
	 */
	message(text);
	synth->clear(synth);
	if (menu->cur == 0)
	    lang->speak_string(lang, text);
	else
	    lang->speak_string(lang, text + mlen);

	event = get_input_event(0, EVNT_BLOCKING);
	switch (event) {
	case CMD_CONFMENU:
	    menu->cur = 0;
	    node  = menu->entry[menu->cur];
				/* no break, fall through */
	case CMD_HWINRT:
	    switch (node->type) {
	    case N_MENU:
		menu = &node->m; /* enter menu */
		break;
	    case N_ENTRY:
		node->e.fun(node->e.private, C_UP, &title, &val);
		break;
	    }
	    break;
	case CMD_HWINLT:
	    switch (node->type) {
	    case N_MENU:
		menu = &node->m; /* enter menu */
		break;
	    case N_ENTRY:
		node->e.fun(node->e.private, C_DOWN, &title, &val);
		break;
	    }
	    break;
	case CMD_LNUP:
	    menu->cur--;
	    if (menu->cur < 0)
		menu->cur = 0;
	    break;
	case CMD_LNDN:
	    menu->cur++;
	    if (menu->cur >= menu->no_entries)
		menu->cur = menu->no_entries - 1;
	    break;
	default:
	    ;			/* ignore unknown event */
	    break;
	}
    }

}


/*
 * ----------------------------------------------------------------------
 * Save the subtree into file.
 * ----------------------------------------------------------------------
 */
static int save_tree(FILE *fp, menu_t *menu, int level, int lang)
{
    node_t *node;
    msg_t  title;
    int    e, i;
    int    val;

    for (i = 0; i < 2 * level; i++)
	fputc(' ', fp);
    fprintf(fp, "%s {\n", menu->title->txt[lang]);
    level++;

    for (e = 1; e < menu->no_entries; e++) {
	node  = menu->entry[e];
	switch (node->type) {
	case N_MENU:
	    save_tree(fp, &node->m, level, lang);
	    break;
	case N_ENTRY:
	    for (i = 0; i < 2 * level; i++)
		fputc(' ', fp);
	    node->e.fun(node->e.private, C_GET_VALUE, &title, &val);
	    fprintf(fp, "%d %s\n", val, title.txt[lang]);
	    break;
	}
    }
    level--;
    for (i = 0; i < 2 * level; i++)
	fputc(' ', fp);
    fprintf(fp, "}\n");

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
int save_setup(const char *filename)
{
    FILE *fp;
    int res = 0;

    fp = fopen(filename, "w");
    if (fp == NULL)
	return 1;

    res = save_tree(fp, root, 0, 0);
    fclose(fp);

    return res;
}


/*
 * ----------------------------------------------------------------------
 * Read and throw away all characters up to and including next '\n'
 * Return 0 on success.
 * ----------------------------------------------------------------------
 */
static int skip_to_eol(FILE *fp)
{
    int ch;

    do {
	ch = fgetc(fp);
    } while (ch != '\n' && ch != EOF);

    return ch == EOF ? 1: 0;
}


/*
 * ----------------------------------------------------------------------
 * Load the subtree from file.
 * ----------------------------------------------------------------------
 */
static int load_tree(FILE *fp, menu_t *menu, int lang)
{
    node_t *node;
    msg_t  title;
    int    e;
    int    val;

    if (skip_to_eol(fp))	/* "menu_title {\n" */
	return 1;

    for (e = 1; e < menu->no_entries; e++) {
	node  = menu->entry[e];
	switch (node->type) {
	case N_MENU:
	    if (load_tree(fp, &node->m, lang))
		return 1;
	    break;
	case N_ENTRY:
	    if (fscanf(fp, "%d", &val) == EOF)
		return 1;
	    node->e.fun(node->e.private, C_SET_VALUE, &title, &val);
	    if (skip_to_eol(fp))
		return 1;
	    break;
	}
    }
    if (skip_to_eol(fp))	/* "}\n" */
	return 1;

    return 0;
}


/*
 * ----------------------------------------------------------------------
 * Load setup from file. On success return 0, else error code
 * (currently only 1).
 * ----------------------------------------------------------------------
 */
int load_setup(const char *filename)
{
    FILE *fp;
    int res = 0;

    fp = fopen(filename, "r");
    if (fp == NULL)
	return 1;

    res = load_tree(fp, root, 0);
    fclose(fp);

    return res;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
menu_t *make_menu(msg_t *title, node_t *parent)
{
    menu_t *menu;

    menu = malloc(sizeof(menu_t));
    if (menu == NULL)
	return NULL;

    menu->type          = N_MENU;
    menu->title         = title;
    menu->no_entries    =  1;
    menu->alloc_entries = 10;
    menu->entry = malloc(10 * sizeof(node_t*));
    menu->cur = 0;
    menu->entry[0] = parent;

    return menu;
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
void add_bin_to_menu(menu_t *menu, msg_t *title, int *value,
		     msg_t *zero_text, msg_t *one_text)
{
    binary_t *bin;
    
    bin = malloc(sizeof(binary_t));
    assert(bin != NULL);
    bin->entry = malloc(sizeof(entry_t));
    assert(bin->entry != NULL);
    bin->value = value;
    bin->text[0] = zero_text;
    bin->text[1] = one_text;
    bin->entry->type = N_ENTRY;
    bin->entry->name = title;
    bin->entry->fun = binary_fun;
    bin->entry->private = bin;
    add_to_menu(menu, (node_t*)bin->entry);
}


/*
 * ----------------------------------------------------------------------
 * 
 * ----------------------------------------------------------------------
 */
void add_itg_to_menu(menu_t *menu, msg_t *title, int *value,
		     int min, int max, int wrap)
{
    integer_t *itg;
    
    itg = malloc(sizeof(integer_t));
    assert(itg != NULL);
    itg->entry = malloc(sizeof(entry_t));
    assert(itg->entry != NULL);
    itg->value = value;
    itg->min   = min;
    itg->max   = max;
    itg->wrap  = wrap;
    itg->entry->type = N_ENTRY;
    itg->entry->name = title;
    itg->entry->fun = integer_fun;
    itg->entry->private = itg;
    add_to_menu(menu, (node_t*)itg->entry);
}


/*
 * ----------------------------------------------------------------------
 * Add a node (entry or menu) to an existing menu,
 * resize menu entry table if necessary.
 * ----------------------------------------------------------------------
 */
void add_to_menu(menu_t *menu, node_t *node)
{
    if (menu->no_entries == menu->alloc_entries)
	menu->alloc_entries += 10;
    menu->entry = realloc(menu->entry, menu->alloc_entries * sizeof(node_t*));
    assert(menu->entry != NULL);
    menu->entry[menu->no_entries++] = node;
}


/*
 * ----------------------------------------------------------------------
 * Dump the subtree into file.
 * ----------------------------------------------------------------------
 */
static void dump_tree(FILE *fp, menu_t *menu, int level, int lang)
{
    node_t *node;
    msg_t  title;
    int    e, i;
    int    val;

    for (i = 0; i < 2 * level; i++)
	fputc(' ', fp);
    fprintf(fp, "%s ", menu->title->txt[lang]);
    node  = menu->entry[0];
    switch (node->type) {
    case N_MENU:
	fprintf(fp, "<%s>", node->m.title->txt[lang]);
	break;
    case N_ENTRY:
	fprintf(fp, "<%s>", node->e.name->txt[lang]);
	break;
    }
    
    fprintf(fp, " {\n");
    level++;

    for (e = 1; e < menu->no_entries; e++) {
	node  = menu->entry[e];
	switch (node->type) {
	case N_MENU:
	    dump_tree(fp, &node->m, level, lang);
	    break;
	case N_ENTRY:
	    for (i = 0; i < 2 * level; i++)
		fputc(' ', fp);
	    node->e.fun(node->e.private, C_GET_VALUE, &title, &val);
	    fprintf(fp, "%s\n", title.txt[lang]);
	    break;
	}
    }
    level--;
    for (i = 0; i < 2 * level; i++)
	fputc(' ', fp);
    fprintf(fp, "}\n");
}


/*
 * ----------------------------------------------------------------------
 * This function is entered into the first slot of the root menu because
 * there is no parent menu for this one.
 * On "up", exit setup system.
 * ----------------------------------------------------------------------
 */
static void exit_fun(void *this, cmd_e cmd, msg_t *text, int *value)
{
    static msg_t t = {
	{ "exit", "verlassen" }
    };

    switch (cmd) {
    case C_GET_VALUE:
	if (value != NULL)
	    *value = 0;
	if (text != NULL)
	    *text = t;
	break;
    case C_SET_VALUE:
	printf("set value\n");
	break;
    case C_UP:
	keep_going = 0;
	menu = root;
	menu->cur = 0;
	break;
    case C_DOWN:
	printf("down\n");
	break;
    }
}


/*
 * ----------------------------------------------------------------------
 * General function for handling of binary values (0, 1).
 * ----------------------------------------------------------------------
 */
static void binary_fun(void *this, cmd_e cmd, msg_t *text, int *value)
{
    binary_t *bin = this;
    int      l;

    switch (cmd) {
    case C_GET_VALUE:
	if (value != NULL)
	    *value = *bin->value;
	break;
    case C_SET_VALUE:
	*bin->value = *value;
	break;
    case C_UP:
    case C_DOWN:
	*bin->value = !*bin->value;
	if (value != NULL)
	    *value = *bin->value;
	break;
    }

    if (text != NULL) {
	for (l = 0; l < SETUP_LANGUAGES; l++) {
	    sprintf(text->txt[l], "%s %s",
		    bin->entry->name->txt[l],
		    bin->text[*value]->txt[l]);
	    
	}
    }
}


/*
 * ----------------------------------------------------------------------
 * General function for handling of inter values (itg->min, itg->max).
 * ----------------------------------------------------------------------
 */
static void integer_fun(void *this, cmd_e cmd, msg_t *text, int *value)
{
    integer_t *itg = this;
    int       l;

    switch (cmd) {
    case C_GET_VALUE:
	if (value != NULL)
	    *value = *itg->value;
	break;
    case C_SET_VALUE:
	*itg->value = *value;
	break;
    case C_UP:
	(*itg->value)++;
	if (*itg->value > itg->max)
	    if (itg->wrap)
		*itg->value = itg->min;
	    else
		*itg->value = itg->max;
	if (value != NULL)
	    *value = *itg->value;
	break;
    case C_DOWN:
	(*itg->value)--;
	if (*itg->value < itg->min)
	    if (itg->wrap)
		*itg->value = itg->max;
	    else
		*itg->value = itg->min;
	if (value != NULL)
	    *value = *itg->value;
	break;
    }

    if (text != NULL) {
	for (l = 0; l < SETUP_LANGUAGES; l++) {
	    sprintf(text->txt[l],
		    "%s %d", itg->entry->name->txt[l], *itg->value);
	    
	}
    }
}


/*
 * ----------------------------------------------------------------------
 * Make the menu to setup braille output parameters.
 * ----------------------------------------------------------------------
 */
static node_t *make_braille_menu(node_t *parent)
{
    static msg_t brl_title = { { "Braille", "Braille" } };
    static msg_t cur_vis   = { { "curosor visible", "Cursor sichtbar" } };
    static msg_t attrvis   = { { "attributes visible", "Attributanzeige" } };
    static msg_t csrblink  = { { "cursor blinking", "Cursor blinkend" } };
    static msg_t capblink  = { { "caps blinking", "Gro blinkend" } };
    static msg_t attrblink = { { "attributes blinking", "Attribute blinkend" } };
    static msg_t csrsize   = { { "cursorsize", "Cursorgre" } };
    static msg_t slidewin  = { { "sliding window", "Gleitendes Fenster" } };
    static msg_t line_text  = { { "underline", "Strich" } };
    static msg_t block_text = { { "block", "Voll" } };
    static msg_t csron_text   = { { "cursor on", "Cursor ein" } };
    static msg_t csroff_text  = { { "cursor off", "Cursor aus" } };
    static msg_t capon_text   = { { "caps on", "Caps ein" } };
    static msg_t capoff_text  = { { "caps off", "Caps aus" } };
    static msg_t attron_text  = { { "attribute on", "Attribute ein" } };
    static msg_t attroff_text = { { "attribute off", "Attribute aus" } };
    menu_t *menu;

    menu = make_menu(&brl_title, parent);
    add_bin_to_menu(menu, &cur_vis, &s_csrvis, &no_text, &yes_text);
    add_bin_to_menu(menu, &attrvis, &s_attrvis, &off_text, &on_text);
    add_bin_to_menu(menu, &csrblink, &s_csrblink, &no_text, &yes_text);
    add_bin_to_menu(menu, &capblink, &s_capblink, &no_text, &yes_text);
    add_bin_to_menu(menu, &attrblink, &s_attrblink, &no_text, &yes_text);
    add_bin_to_menu(menu, &csrsize, &s_csrsize, &line_text, &block_text);
    add_bin_to_menu(menu, &slidewin, &s_slidewin, &no_text, &yes_text);

    add_itg_to_menu(menu, &csron_text, &s_csroncnt, 1, 16, 1);
    add_itg_to_menu(menu, &csroff_text, &s_csroffcnt, 1, 16, 1);
    add_itg_to_menu(menu, &capon_text, &s_caponcnt, 1, 16, 1);
    add_itg_to_menu(menu, &capoff_text, &s_capoffcnt, 1, 16, 1);
    add_itg_to_menu(menu, &attron_text, &s_attroncnt, 1, 16, 1);
    add_itg_to_menu(menu, &attroff_text, &s_attroffcnt, 1, 16, 1);

    return (node_t*)menu;
}


/*
 * ----------------------------------------------------------------------
 * Set language.
 * ----------------------------------------------------------------------
 */
static void language_fun(void *this, cmd_e cmd, msg_t *text, int *value)
{
    language_t *lg = this;
    int         l;
    int         scr;

    if (lg == NULL) {
	scr = curscr;
	lg = &langs[scr];
	assert(lg != NULL);
    } else {
	scr = lg - langs;
    }
    switch (cmd) {
    case C_GET_VALUE:
	if (value != NULL)
	    *value = lg->langi;
	break;
    case C_SET_VALUE:
	lg->langi = *value;
	break;
    case C_UP:
	lg->langi++;
	if (lg->langi >= SETUP_LANGUAGES)
	    lg->langi = 0;
	if (value != NULL)
	    *value = lg->langi;
	break;
    case C_DOWN:
	lg->langi--;
	if (lg->langi < 0)
	    lg->langi = SETUP_LANGUAGES - 1;
	if (value != NULL)
	    *value = lg->langi;
	break;
    }
    /*
     * Set the language in screen structure.
     */
    scrparam[scr].speech = &speech_chan[lg->langi];
    
    if (text != NULL) {
	for (l = 0; l < SETUP_LANGUAGES; l++) {
	    sprintf(text->txt[l],
		    "%s %d %s", lg->entry->name->txt[l],
		    lg - &langs[0],
		    lang_names[lg->langi].txt[l]);
	}
    }
}


/*
 * ----------------------------------------------------------------------
 * Set a parameter of a synthesizer or language object.
 * ----------------------------------------------------------------------
 */
static void speech_par_fun(void *this, cmd_e cmd, msg_t *text, int *value)
{
    speech_par_t *par = this;
    int          l;
    int          v;

    switch (cmd) {
    case C_GET_VALUE:
	if (par->is_lang)
	    par->u.lang->get_param(par->u.lang, par->par, &v);
	else
	    par->u.synth->get_param(par->u.synth, par->par, &v);
	if (value != NULL)
	    *value = v;
	break;
    case C_SET_VALUE:
	v = *value;
	if (par->is_lang)
	    par->u.lang->set_param(par->u.lang, par->par, v);
	else
	    par->u.synth->set_param(par->u.synth, par->par, v);
	break;
    case C_UP:
	if (par->is_lang)
	    par->u.lang->get_param(par->u.lang, par->par, &v);
	else
	    par->u.synth->get_param(par->u.synth, par->par, &v);
	if (v < 15)
	    v++;		/* increment by one */
	else
	    v += 0.10 * v;	/* add 10% */
	if (par->is_lang)
	    par->u.lang->set_param(par->u.lang, par->par, v);
	else
	    par->u.synth->set_param(par->u.synth, par->par, v);
	*value = v;
	break;
    case C_DOWN:
	if (par->is_lang)
	    par->u.lang->get_param(par->u.lang, par->par, &v);
	else
	    par->u.synth->get_param(par->u.synth, par->par, &v);
	if (v <= 15)
	    v--;		/* decrement by one */
	else
	    v -= 0.10 * v;	/* subtract 10% */
	if (v < 0)
	    v = 0;
	if (par->is_lang)
	    par->u.lang->set_param(par->u.lang, par->par, v);
	else
	    par->u.synth->set_param(par->u.synth, par->par, v);
	*value = v;
	break;
    }
    
    if (text != NULL) {
	for (l = 0; l < SETUP_LANGUAGES; l++) {
	    sprintf(text->txt[l],
		    "%s %d", par->entry->name->txt[l], v);
	}
    }
}


/*
 * ----------------------------------------------------------------------
 * Make the menu to setup language parameters.
 * ----------------------------------------------------------------------
 */
static node_t *make_language_menu(node_t *parent, int langi)
{
    static msg_t l_speed  = { { "speed", "Geschwindigkeit" } };
    static msg_t l_volume = { { "volume", "Lautstrke" } };
    static msg_t l_pitch  = { { "pitch", "Tonhhe" } };
    static msg_t l_punct  = { { "punctuation", "Satzzeichen" } };
    entry_t      *entry;
    speech_par_t *par;
    menu_t       *menu;

    menu = make_menu(&lang_names[langi], parent);
    /*
     * Entries for speed, volume, pitch. For all of them do:
     */
    entry = malloc(sizeof(entry_t));
    assert(entry != NULL);
    par = malloc(sizeof(speech_par_t));
    assert(par != NULL);
    entry->private = par;
    entry->type    = N_ENTRY;
    entry->name    = &l_speed;
    entry->fun     = speech_par_fun;
    par->entry   = entry;
    par->is_lang = 0;
    par->u.synth = speech_chan[langi].lang->synth;
    par->par     = S_SPEED;
    add_to_menu(menu, (node_t*)entry);
    
    entry = malloc(sizeof(entry_t));
    assert(entry != NULL);
    par = malloc(sizeof(speech_par_t));
    assert(par != NULL);
    entry->private = par;
    entry->type    = N_ENTRY;
    entry->name    = &l_volume;
    entry->fun     = speech_par_fun;
    par->entry   = entry;
    par->is_lang = 0;
    par->u.synth = speech_chan[langi].lang->synth;
    par->par     = S_VOLUME;
    add_to_menu(menu, (node_t*)entry);
    
    entry = malloc(sizeof(entry_t));
    assert(entry != NULL);
    par = malloc(sizeof(speech_par_t));
    assert(par != NULL);
    entry->private = par;
    entry->type    = N_ENTRY;
    entry->name    = &l_pitch;
    entry->fun     = speech_par_fun;
    par->entry   = entry;
    par->is_lang = 0;
    par->u.synth = speech_chan[langi].lang->synth;
    par->par     = S_PITCH;
    add_to_menu(menu, (node_t*)entry);
    
    entry = malloc(sizeof(entry_t));
    assert(entry != NULL);
    par = malloc(sizeof(speech_par_t));
    assert(par != NULL);
    entry->private = par;
    entry->type    = N_ENTRY;
    entry->name    = &l_punct;
    entry->fun     = speech_par_fun;
    par->entry   = entry;
    par->is_lang = 1;
    par->u.lang  = speech_chan[langi].lang;
    par->par     = L_SPEAK_PUNCTUATION;
    add_to_menu(menu, (node_t*)entry);
    

    return (node_t*)menu;
}


/*
 * ----------------------------------------------------------------------
 * Make the menu to setup speech parameters.
 * ----------------------------------------------------------------------
 */
static node_t *make_speech_menu(node_t *parent)
{
    static msg_t title = { { "Speech", "Sprachausgabe" } };
    static msg_t sp_on   = { { "Speech", "Sprache" } };
    static msg_t screen  = { { "Terminal", "Bildschirm" } };
    entry_t *entry;
    menu_t  *menu;
    int     i;

    menu = make_menu(&title, parent);
    /*
     * Top level entry: Language of current screen.
     */
    entry = malloc(sizeof(entry_t));
    assert(entry != NULL);
    entry->type = N_ENTRY;
    entry->name = &screen;
    entry->fun  = language_fun;
    
    langs[0].langi = 0;
    langs[0].p     = &scrparam[i];
    langs[0].entry = entry;

    entry->private = NULL;
    add_to_menu(menu, (node_t*)entry);
    /*
     * General switch for speech on/off.
     */
    add_bin_to_menu(menu, &sp_on, &s_speech, &off_text, &on_text);
    /*
     * Parameter menus for all languages.
     */
    for (i = 0; i < SETUP_LANGUAGES; i++) {
	add_to_menu(menu, make_language_menu((node_t*)menu, i));
    }
    /*
     * And at the end: The long list with the language for all screens.
     */
    for (i = 1; i < NBR_SCR; i++) {
	entry = malloc(sizeof(entry_t));
	assert(entry != NULL);
	entry->type = N_ENTRY;
	entry->name = &screen;
	entry->fun  = language_fun;

	langs[i].langi = 0;
	langs[i].p     = &scrparam[i];
	langs[i].entry = entry;

	entry->private = &langs[i];
	add_to_menu(menu, (node_t*)entry);
    }

    return (node_t*)menu;
}
